import os
import shutil
from collections.abc import Generator
from pathlib import Path
from textwrap import dedent
from typing import Any

import ddeutil.io.__base.files as fl
import pytest
import yaml


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    target_path: Path = test_path / "base_file_yaml"
    target_path.mkdir(parents=True, exist_ok=True)

    yield target_path

    shutil.rmtree(target_path)


@pytest.fixture(scope="module")
def yaml_str_safe() -> str:
    return dedent(
        """
    main_key:
        sub_key:
            string: 'test ${DEMO_ENV_VALUE} value'
            int: 0.001
            bool: false
            list: ['i1', 'i2', 'i3']
            str2bool: on
            statement: |
                # Comment ${DEMO_ENV_VALUE}
                This is a long statement with comment above
    """
    ).strip()


@pytest.fixture(scope="module")
def yaml_data_safe() -> dict[str, Any]:
    return {
        "main_key": {
            "sub_key": {
                "string": "test ${DEMO_ENV_VALUE} value",
                "int": 0.001,
                "bool": False,
                "list": ["i1", "i2", "i3"],
                "str2bool": True,
                "statement": (
                    "# Comment ${DEMO_ENV_VALUE}\n"
                    "This is a long statement with comment above"
                ),
            }
        }
    }


def test_write_yaml_file_with_safe(target_path, yaml_str_safe, yaml_data_safe):
    yaml_path: Path = target_path / "test_write_file.yaml"
    fl.YamlFl(path=yaml_path).write(yaml_data_safe)
    assert yaml_path.exists()


def test_read_yaml_resolve_file(target_path, yaml_str_safe, yaml_data_safe):
    yaml_path: Path = target_path / "test_read_file_resolve.yaml"
    with open(yaml_path, mode="w", encoding="utf-8") as f:
        f.write(yaml_str_safe)

    data = fl.YamlFlResolve(path=yaml_path).read(safe=False)
    assert "on" == data["main_key"]["sub_key"]["str2bool"]
    assert (
        "# Comment ${DEMO_ENV_VALUE}\n"
        "This is a long statement with comment above"
    ) == data["main_key"]["sub_key"]["statement"]


def test_read_yaml_file_with_safe(target_path, yaml_str_safe, yaml_data_safe):
    yaml_path: Path = target_path / "test_read_file_safe.yaml"
    with open(yaml_path, mode="w", encoding="utf-8") as f:
        yaml.dump(yaml.safe_load(yaml_str_safe), f)

    data = fl.YamlFl(path=yaml_path).read()
    assert yaml_data_safe == data


def test_read_yaml_file(target_path, yaml_str_safe, yaml_data_safe):
    yaml_path: Path = target_path / "test_read_file.yaml"
    with open(yaml_path, mode="w", encoding="utf-8") as f:
        f.write(yaml_str_safe)

    data = fl.YamlFl(path=yaml_path).read(safe=False)
    assert yaml_data_safe == data


@pytest.fixture(scope="module")
def yaml_str_env_safe() -> str:
    return dedent(
        """
    main_key:
        sub_key:
            key01: 'test ${DEMO_ENV_VALUE} value'
            key02: $1 This is escape with number
            key03: $$ESCAPE This is escape with $
            key04: ['i1', 'i2', '${DEMO_ENV_VALUE}']
            key05: ${DEMO_ENV_VALUE_EMPTY:default}
            key06: $${DEMO_ENV_VALUE}
            key07: This ${DEMO_ENV_VALUE} ${{DEMO_ENV_VALUE}}
            key08: |
                # Comment
                statement ${DEMO_ENV_VALUE}
    """
    ).strip()


@pytest.fixture(scope="module")
def yaml_data_env_safe() -> dict[str, Any]:
    return {
        "main_key": {
            "sub_key": {
                "key01": "test demo value",
                "key02": "$1 This is escape with number",
                "key03": "$ESCAPE This is escape with $",
                "key04": ["i1", "i2", "demo"],
                "key05": "default",
                "key06": "${DEMO_ENV_VALUE}",
                "key07": "This demo ${{DEMO_ENV_VALUE}}",
                "key08": "# Comment\nstatement demo",
            },
        },
    }


def test_read_yaml_file_with_safe_mode(
    yaml_str_env_safe,
    yaml_data_env_safe,
    target_path,
):
    yaml_path: str = target_path / "test_read_file_env.yaml"

    with open(yaml_path, mode="w", encoding="utf-8") as f:
        yaml.dump(yaml.safe_load(yaml_str_env_safe), f)

    os.environ["DEMO_ENV_VALUE"] = "demo"

    data = fl.YamlEnvFl(path=yaml_path).read()
    assert yaml_data_env_safe == data


def test_read_yaml_file_with_safe_mode_and_prepare(
    target_path,
    yaml_str_env_safe,
):
    yaml_path: str = target_path / "test_read_file_env_prepare.yaml"

    with open(yaml_path, mode="w", encoding="utf-8") as f:
        yaml.dump(yaml.safe_load(yaml_str_env_safe), f)

    os.environ["DEMO_ENV_VALUE"] = "demo"

    yml_loader = fl.YamlEnvFl(path=yaml_path)
    yml_loader.prepare = lambda x: f"{x}!!"
    data = yml_loader.read()
    assert {
        "main_key": {
            "sub_key": {
                "key01": "test demo!! value",
                "key02": "$1 This is escape with number",
                "key03": "$ESCAPE This is escape with $",
                "key04": ["i1", "i2", "demo!!"],
                "key05": "default!!",
                "key06": "${DEMO_ENV_VALUE}",
                "key07": "This demo!! ${{DEMO_ENV_VALUE}}",
                "key08": "# Comment\nstatement demo!!",
            }
        }
    } == data


def test_read_yaml_file_with_safe_mode_and_prepare_2(
    target_path,
    yaml_str_env_safe,
):
    yaml_path: str = target_path / "test_read_file_env_prepare_2.yaml"

    with open(yaml_path, mode="w", encoding="utf-8") as f:
        yaml.dump(yaml.safe_load(yaml_str_env_safe), f)

    os.environ["DEMO_ENV_VALUE"] = "P@ssW0rd"

    import urllib.parse

    yml_loader = fl.YamlEnvFl
    yml_loader.prepare = staticmethod(lambda x: urllib.parse.quote_plus(str(x)))
    data = yml_loader(path=yaml_path).read()
    assert {
        "main_key": {
            "sub_key": {
                "key01": "test P%40ssW0rd value",
                "key02": "$1 This is escape with number",
                "key03": "$ESCAPE This is escape with $",
                "key04": ["i1", "i2", "P%40ssW0rd"],
                "key05": "default",
                "key06": "${DEMO_ENV_VALUE}",
                "key07": "This P%40ssW0rd ${{DEMO_ENV_VALUE}}",
                "key08": "# Comment\nstatement P%40ssW0rd",
            }
        }
    } == data



import os

import ddeutil.io.__base.utils as utils
import pytest


def test_search_env_replace():
    os.environ["NAME"] = "foo"
    rs = utils.search_env_replace("Hello ${NAME}")
    assert "Hello foo" == rs


def test_search_env_replace_raise():
    with pytest.raises(ValueError):
        utils.search_env_replace(
            "Hello ${NAME01}", raise_if_default_not_exists=True
        )

    with pytest.raises(ValueError):
        utils.search_env_replace("Hello ${:test}")


def test_search_env():
    rs = utils.search_env(
        ("key='demo'\n" "# foo=bar\n" "hello=${key}-2\n" "escape=\\${key}\n"),
    )
    assert {"key": "demo", "hello": "demo-2", "escape": "${key}"} == rs


def test_search_env_raise():
    with pytest.raises(ValueError):
        utils.search_env("foo=")

    with pytest.raises(ValueError):
        utils.search_env("foo=''")

    with pytest.raises(ValueError):
        utils.search_env('foo=""')



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.__base.files as fl
import pytest


@pytest.fixture(scope="module")
def csv_path(test_path) -> Generator[Path, None, None]:
    this_path: Path = test_path / "open_file"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_csv_read_and_write(csv_path):
    csv_data: list[str] = [
        {"Col01": "A", "Col02": "1", "Col03": "test1"},
        {"Col01": "B", "Col02": "2", "Col03": "test2"},
        {"Col01": "C", "Col02": "3", "Col03": "test3"},
    ]
    fl.CsvFl(csv_path / "test_file.csv").write(csv_data)
    assert csv_data == fl.CsvFl(csv_path / "test_file.csv").read()



from .__about__ import __version__
from .__base import (
    PathSearch,
    RegexConf,
    rm,
    touch,
)
from .__base.files import (
    CsvFl,
    CsvPipeFl,
    EnvFl,
    Fl,
    JsonEnvFl,
    JsonFl,
    MarshalFl,
    PickleFl,
    TomlEnvFl,
    TomlFl,
    YamlEnvFl,
    YamlFl,
    YamlFlResolve,
)
from .__base.utils import (
    search_env,
    search_env_replace,
)
from .config import (
    ConfABC,
    ConfFl,
    ConfSQLite,
)
from .exceptions import (
    ConfigArgumentError,
    ConfigNotFound,
    IOBaseError,
)
from .param import (
    Engine,
    Params,
    PathData,
    Rule,
    Stage,
    Value,
)
from .register import Register
from .utils import (
    map_func,
    map_importer,
    map_secret,
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import re
from typing import (
    Optional,
    Union,
)
from urllib.parse import unquote_plus


def unquote_str(value: str):
    """
    :param value:
    :return:

    Examples:
        >>> unquote_str('P%40ssw0rd')
        'P@ssw0rd'
        >>> unquote_str(None)

    """
    return unquote_plus(value) if value else value


def catch_str(
    value: str,
    key: str,
    *,
    replace: Optional[str] = None,
    flag: bool = True,
) -> tuple[str, Optional[Union[bool, str]]]:
    """Catch keyword from string value and return True if it exits.

    :param value: a string value that want to catch with the key.
    :param key: a key that use to catch.
    :param replace: a string value that want to replace if catch found
    :param flag: if it true, it will return boolean instead catched string

    :rtype: tuple[str, Optional[Union[bool, str]]]
    :return: a pair or prepared value and catched string or boolean if flag wan
        setted.

    Examples:
        >>> catch_str("varchar( 100 ) unique", "unique")
        ('varchar( 100 )', True)
        >>> catch_str("integer primary keys", "primary key")
        ('integer s', True)
        >>> catch_str("timestamp( 6 ) not null", "not null", replace="null")
        ('timestamp( 6 ) null', True)
        >>> catch_str("integer primary keys", "primary key", flag=False)
        ('integer s', 'primary key')
    """
    if key in value:
        return (
            " ".join(value.replace(key, (replace or "")).split()),
            (True if flag else key),
        )
    return value, (False if flag else None)


def split_dtype(dtype: str) -> tuple[str, str]:
    """Split the datatype value from long string by null string

    :param dtype: a data type string value that want to split
    :type dtype: str

    :rtype: tuple[str, str]
    :return: a pair of full data type value and null/not null string value.

    Examples:
        >>> split_dtype("string null")
        ('string', 'null')
        >>> split_dtype("numeric(10, 2) Null")
        ('numeric(10, 2)', 'null')
        >>> split_dtype("timestamp(6) NULL")
        ('timestamp(6)', 'null')
        >>> split_dtype("string not null")
        ('string', 'not null')
        >>> split_dtype("varchar( 20 ) not null null")
        ('varchar( 20 )', 'null')
        >>> split_dtype("string null null")
        ('string', 'null')
    """
    _nullable: str = "null"
    for null_str in (
        "not null",
        "Not Null",
        "NOT NULL",
        "null",
        "Null",
        "NULL",
    ):
        if re.search(null_str, dtype):
            _nullable = null_str
            dtype = dtype.replace(null_str, "")
    return " ".join(dtype.strip().split()), _nullable.lower()


def extract_dtype(dtype: str) -> dict[str, str]:
    """Extract Data Type from an input string.

    Examples:
        >>> extract_dtype("varchar( 255 )")
        {'type': 'varchar', 'max_length': '255'}
        >>> extract_dtype("varchar[10]")
        {'type': 'varchar', 'max_length': '10'}
        >>> extract_dtype("numeric(19, 2)")
        {'type': 'numeric', 'precision': '19', 'scale': '2'}
    """
    if m := re.search(
        r"(?P<type>\w+)"
        r"(?:\s?[(\[]\s?(?P<max_length>\d+)(?:,\s?(?P<scale>\d+))?\s?[])])?",
        dtype.strip(),
    ):
        extract = m.groupdict()
        if (t := extract["type"]) in ("numeric", "decimal"):
            extract["precision"] = extract.pop("max_length")
            extract["scale"] = extract.pop("scale", None) or -1
            return extract

        extract.pop("scale")
        if t in ("timestamp", "time"):
            extract["precision"] = extract.pop("max_length")
            return extract
        return extract
    return {"type": dtype}



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from typing import Annotated, Optional, Union

from pydantic import (
    Field,
    ValidationInfo,
)
from pydantic.functional_validators import field_validator
from pydantic.types import SecretStr

from ..__base import BaseUpdatableModel
from ..const import Fk, Pk
from .col import Col


class BaseDef(BaseUpdatableModel):
    """Base Definition"""

    name: Annotated[
        str,
        Field(description="Name"),
    ]
    desc: Annotated[
        Optional[str],
        Field(default=None, description="Description of Object"),
    ]


class BaseTbl(BaseDef):
    """Base Table Model"""

    features: Annotated[
        list[Col],
        Field(
            default_factory=list,
            description="Schema of this Table",
        ),
    ]


class Tbl(BaseTbl):
    """Table Model"""

    pk: Annotated[
        Pk,
        Field(validate_default=True, description="Primary Key"),
    ] = Pk()
    fk: Annotated[
        list[Fk],
        Field(default_factory=list, description="Foreign Key"),
    ]

    consts: Annotated[
        list[str],
        Field(default_factory=list),
    ]

    @field_validator("pk")
    def __receive_pk_from_features(cls, value: Pk, info: ValidationInfo) -> Pk:
        """Receive the primary key from the features."""
        # NOTE:
        #   we respect that `info.data` should contain schema before `pk`
        #   validation.
        pks: list[str] = [
            i.name for i in filter(lambda x: x.pk, info.data["features"])
        ]
        if pks and not value.cols:
            # Note: pass primary key cols if `pk` does not set.
            value = Pk(cols=list(pks))

        # NOTE: change name of `pk` with parent Tbl class name
        value.of = info.data["name"]
        return value

    @field_validator("fk")
    def __prepare_fk(cls, value: list[Fk], info: ValidationInfo) -> list[Fk]:
        # NOTE: change name of `fk` with parent Tbl class name
        for fk in value:
            fk.of = info.data["name"]
        return value


class BaseDefine(BaseDef):
    definition: str


class Func(BaseDefine): ...


class Proc(BaseDefine): ...


class View(BaseDefine): ...


class BaseScm(BaseUpdatableModel):
    """Base Schema Model"""

    name: str
    tables: Annotated[
        list[Tbl],
        Field(default_factory=list, description="Collection of Tables"),
    ]


class Scm(BaseScm):
    """Schema Model"""

    views: Annotated[
        list[View],
        Field(default_factory=list, description="Collection of Views"),
    ]
    funcs: Annotated[
        list[Func],
        Field(default_factory=list, description="Collection of Functions"),
    ]
    procs: Annotated[
        list[Proc],
        Field(default_factory=list, description="Collection of Procedures"),
    ]


class BaseUser(BaseUpdatableModel):
    name: str


class SQLUser(BaseUser):
    pwd: Annotated[Optional[SecretStr], Field(description="Password")] = None


class Role(BaseUpdatableModel):
    name: str
    members: list[SQLUser] = Field(default_factory=list)


User = Union[SQLUser, BaseUser]


class BaseDb(BaseUpdatableModel):
    schemas: Annotated[
        list[Scm],
        Field(default_factory=list, description="Collection of Schemas"),
    ]


class Db(BaseDb):
    users: list[User]
    roles: list[Role]
    policies: list[str]



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import date, datetime, timezone
from typing import (
    Annotated,
    Any,
    Optional,
)
from zoneinfo import ZoneInfo

from pydantic import BaseModel, Field
from pydantic.functional_validators import field_validator, model_validator

try:
    from .__base import BaseUpdatableModel
    from .__enums import Status
except ImportError:
    from __base import BaseUpdatableModel
    from __enums import Status


TZ: str = "Asia/Bangkok"


def dt_now() -> datetime:
    return datetime.now(timezone.utc)


class Ts(BaseModel):
    """Timestamp Model

    Examples:
        >>> from datetime import datetime
        >>> Ts(ts=datetime(2024, 1, 1, 5)).ts.tzinfo
        zoneinfo.ZoneInfo(key='Asia/Bangkok')
    """

    ts: Annotated[datetime, Field(default_factory=dt_now, alias="Timestamp")]
    tz: Annotated[str, Field(alias="TimeZone")] = TZ

    @model_validator(mode="after")
    def __prepare_time(self):
        self.ts: datetime = self.ts.astimezone(ZoneInfo(self.tz))
        return self

    def now(self) -> datetime:
        """Return updated timestamp"""
        return datetime.now(tz=self.tz)


class Tag(Ts):
    """Tag Model"""

    author: Annotated[
        Optional[str],
        Field(validate_default=True, description="Author"),
    ] = None
    desc: Annotated[
        Optional[str],
        Field(repr=False, description="Description"),
    ] = None
    labels: Annotated[
        list[str],
        Field(default_factory=list, description="Labels of Tag"),
    ]
    vs: Annotated[
        Optional[date],
        Field(validate_default=True, description="Version of Tag"),
    ] = None

    @field_validator("author")
    def __set_author(cls, value: str | None):
        return value or "undefined"

    @field_validator("vs")
    def __set_version(cls, value: Optional[date]):
        """Pre initialize the `version` value that parsing from default"""
        return value if value else date(year=1990, month=1, day=1)


class BaseTask(BaseUpdatableModel):
    """Base Task Model"""

    st: Status = Field(default=Status.WAITING, description="Status")
    dt: Annotated[
        datetime,
        Field(default_factory=dt_now, alias="Datetime"),
    ]


class Task(BaseTask): ...


class BaseMsg(BaseUpdatableModel):
    level: int
    msg: str


class Msg(BaseMsg): ...


class Log(BaseUpdatableModel):
    """Log Model"""

    msgs: list[Msg]


class BaseParam(BaseUpdatableModel):
    extras: dict[str, Any]

    @model_validator(mode="before")
    def __prepare_extras(cls, values):
        extras: dict[str, Any] = {
            k: values.pop(k) for k in values.copy() if k not in cls.model_fields
        } | values.pop("extras", {})
        return {
            "extras": extras,
            **values,
        }


class NormalParam(BaseParam):
    run_date: datetime



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.__base.files as fl
import pytest


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    target_path: Path = test_path / "open_dir"
    target_path.mkdir(parents=True, exist_ok=True)

    data_path: Path = test_path / "open_dir/data"
    data_path.mkdir(parents=True, exist_ok=True)

    with open(data_path / "test_file.json", mode="w") as f:
        f.write('{"key": "value"}')

    with open(data_path / "test_file_2.json", mode="w") as f:
        f.write('{"foo": "bar"}')

    yield target_path

    shutil.rmtree(target_path)


@pytest.fixture(scope="module")
def data_path(target_path):
    return target_path / "data"


def test_open_dir_common_zip(target_path, data_path):
    with fl.Dir(
        path=target_path / "test_common_zip.zip",
        compress="zip",
    ).open(mode="w") as d:
        for data in data_path.rglob("*"):
            d.write(filename=data, arcname=data.relative_to(data_path))

    with fl.Dir(
        path=target_path / "test_common_zip.zip",
        compress="zip",
    ).open(mode="r") as d:
        d.safe_extract(target_path / "test_common_zip_extract")

    assert {
        "test_file.json",
        "test_file_2.json",
    } == {f.name for f in (target_path / "test_common_zip_extract").rglob("*")}


def test_open_dir_common_tar(target_path, data_path):
    with fl.Dir(
        path=target_path / "test_common_tar.tar.gz",
        compress="tar",
    ).open(mode="w") as d:
        for data in data_path.rglob("*"):
            d.write(name=data, arcname=data.relative_to(data_path))

    with fl.Dir(
        path=target_path / "test_common_tar.tar.gz",
        compress="tar:gz",
    ).open(mode="r") as d:
        d.safe_extract(target_path / "test_common_tar_extract")

    assert {
        "test_file.json",
        "test_file_2.json",
    } == {f.name for f in (target_path / "test_common_tar_extract").rglob("*")}



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Annotated, Any, Literal, Optional, TypeVar

from pydantic import BaseModel, Field
from pydantic.functional_validators import field_validator

T = TypeVar("T")


class BaseAct(BaseModel):
    type: Annotated[str, Field(description="Activity Name")]
    desc: Annotated[Optional[str], Field(description="Description")] = None
    options: Annotated[dict[str, Any], Field(default_factory=dict)]


class Copy(BaseAct):
    type: Literal["copy"] = "copy"
    src: str
    sink: str


class Loop(BaseAct):
    type: Literal["forloop"] = "loop"
    elements: list[T]
    do: str

    @field_validator("elements")
    def __elements_validator(cls, values: list[T]) -> list[T]:
        if len(values) > 0:
            _first: Any = type(values[0])
            if any(not isinstance(value, _first) for value in values[1:]):
                raise TypeError(
                    "all element in loop activity must be the same type"
                )
        return values


class Until(BaseAct):
    type: Literal["forloop"] = "until"
    condition: str


class If(BaseAct):
    type: Literal["if"] = "if"
    condition: str
    right: str = Field(description="The right for correct with condition")
    left: str = Field(description="The left for wrong with condition")


class Sensor(BaseAct):
    type: Literal["sensor"] = "sensor"
    track: str


class Hook(BaseAct):
    type: Literal["hook"] = "hook"
    hook: str



import ddeutil.io.utils as utils
import pytest


def test_map_secrets():
    rs = utils.map_secret(
        "Value include secrets: s3://@secrets{foo}",
        secrets={"foo": "bar"},
    )
    assert "Value include secrets: s3://bar" == rs

    rs = utils.map_secret(
        {
            "list": ["1", "2", "s3://@secrets{foo}"],
            "dict": {
                "tuple": ("1", "2", "s3://@secrets{foo}"),
                "key": 1,
                "boolean": True,
            },
            "default": "s3://@secrets{test:default}",
        },
        secrets={"foo": "bar"},
    )
    assert {
        "list": ["1", "2", "s3://bar"],
        "dict": {
            "tuple": ("1", "2", "s3://bar"),
            "key": 1,
            "boolean": True,
        },
        "default": "s3://default",
    } == rs


def test_map_secrets_raise():
    with pytest.raises(ValueError):
        utils.map_secret(
            "Value include secrets: s3://@secrets{foo.name}",
            secrets={"foo": "bar"},
        )


def test_map_importer():
    rs = utils.map_importer(
        "Test @function{ddeutil.io.__base.add_newline:'a',newline='|'}"
    )
    assert "Test a|" == rs

    reuse: str = "@function{ddeutil.io.__base.add_newline:'a',newline='|'}"
    rs = utils.map_importer(
        {
            "list": [reuse, 1],
            "tuple": (reuse, 2, 3),
        }
    )
    assert {
        "list": ["a|", 1],
        "tuple": ("a|", 2, 3),
    } == rs


def test_map_importer_raise():
    with pytest.raises(ValueError):
        utils.map_importer("@function{ddeutil.io.__version__:'a'}")


def test_map_func():
    rs = utils.map_func({"foo": "bar"}, lambda x: x + "!")
    assert {"foo": "bar!"} == rs

    rs = utils.map_func(("foo", "bar", 1), lambda x: x + "!")
    assert ("foo!", "bar!", 1) == rs



import shutil
from collections.abc import Generator
from pathlib import Path

import pytest
import yaml
from ddeutil.io.param import Params
from ddeutil.io.register import Register


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    tgt_path: Path = test_path / "register_deploy_temp"
    tgt_path.mkdir(exist_ok=True)
    (tgt_path / "conf/demo").mkdir(parents=True, exist_ok=True)
    with open(tgt_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "conn.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path
    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def params(target_path, root_path) -> Params:
    return Params.model_validate(
        {
            "engine": {
                "paths": {
                    "conf": target_path / "conf",
                    "data": root_path / "data",
                    "archive": root_path / "/data/.archive",
                },
                "flags": {"auto_update": True},
            },
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "staging": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                "persisted": {
                    "format": "{domain:%s}_{naming:%s}.{compress:%-g}",
                    "rules": {
                        "compress": "gzip",
                    },
                },
            },
        }
    )


def test_register_deployment(params):
    rs = Register(name="demo:conn_local_file", params=params).deploy()
    assert "(demo:conn_local_file, persisted)" == str(rs)



# -------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# --------------------------------------------------------------------------
import re
from re import (
    IGNORECASE,
    MULTILINE,
    UNICODE,
    VERBOSE,
    Pattern,
)


class RegexConf:
    """Core Configuration Model"""

    # NOTE: Normal regular expression for the secret value.
    # ---
    # [\"\']?                             # single or double quoted value
    # (?P<search>@secrets{                # search string for replacement
    #     (?P<braced>.*?)                 # value if use braced {}
    #     (?::(?P<braced_default>.*?))?   # value default with sep :
    # })                                  # end with }
    # [\"\']?                             # single or double-quoted value
    #
    # NOTE: For secrets grouping level.
    # ---
    # [\"\']?                             # single or double quoted value
    # (?P<search>@secrets                 # search string for replacement
    # (?P<group>(\.\w+)*)?{               # search groups
    #     (?P<braced>.*?)                 # value if use braced {}
    #     (:(?P<braced_default>.*?))?     # value default with sep :
    # })                                  # end with }
    # [\"\']?                             # single or double-quoted value
    __re_secrets: str = r"""
        [\"\']?                             # single or double quoted value
        (?P<search>@secrets{                # search string for replacement
            (?P<braced>.*?)                 # value if use braced {}
            (?::(?P<braced_default>.*?))?   # value default with sep :
        })                                  # end with }
        [\"\']?                             # single or double-quoted value
    """
    RE_SECRETS: Pattern = re.compile(
        __re_secrets, MULTILINE | UNICODE | IGNORECASE | VERBOSE
    )

    # NOTE: Normal regular expression for the function value.
    # ---
    # [\"\']?                             # single or double quoted value
    # (?P<search>@function{               # search string for replacement
    #     (?P<function>[\w.].*?)          # called function
    #     (?::(?P<arguments>.*?))?        # arguments for calling function
    # })                                  # end with }
    # [\"\']?                             # single or double-quoted value
    __re_function: str = r"""
        [\"\']?                             # single or double quoted value
        (?P<search>@function{               # search string for replacement
            (?P<function>[\w.].*?)          # called function
            (?::(?P<arguments>.*?))?        # arguments for calling function
        })                                  # end with }
        [\"\']?                             # single or double-quoted value
    """
    RE_FUNCTION: Pattern = re.compile(
        __re_function, MULTILINE | UNICODE | IGNORECASE | VERBOSE
    )

    # NOTE: Normal regular expression for dotenv variable
    # ---
    # (\\)?(\$)({?([A-Z0-9_]+)}?)
    __re_dotenv_var: str = r"""
        (\\)?               # is it escaped with a backslash?
        (\$)                # literal $
        (                   # collect braces with var for sub
            {?              # allow brace wrapping
            ([A-Z0-9_]+)    # match the variable
            }?              # closing brace
        )                   # braces end
    """
    RE_DOTENV_VAR: Pattern = re.compile(__re_dotenv_var, IGNORECASE | VERBOSE)

    # NOTE: Normal regular expression for dotenv
    # ---
    # ^\s*(?:export\s+)?(?P<name>[\w.-]+)(?:\s*=\s*?|:\s+?)(?P<value>\s*\'(?:\\'|[^'])*\'|\s*\"(?:\\"|[^"])*\"
    # |\s*`(?:\\`|[^`])*`|[^#\r\n]+)?\s*$
    __re_dotenv: str = r"""
        ^\s*(?:export\s+)?          # optional export
        (?P<name>[\w.-]+)           # name of key
        (?:\s*=\s*?|:\s+?)          # separator `=` or `:`
        (?P<value>
            \s*\'(?:\\'|[^'])*\'    # single quoted value
            |
            \s*\"(?:\\"|[^"])*\"    # double quoted value
            |
            \s*`(?:\\`|[^`])*`      # backticks value
            |
            [^#\r\n]+               # unquoted value
        )?\s*                       # optional space
        (?:[^\S\r\n]*\#[^\r\n]*)?
        $
    """
    RE_DOTENV: Pattern = re.compile(__re_dotenv, MULTILINE | VERBOSE)

    # NOTE:
    # ---
    # (\s|^)#.*
    __re_yaml_comment: str = r"(\s|^)#.*"
    RE_YAML_COMMENT: Pattern = re.compile(
        __re_yaml_comment, MULTILINE | UNICODE | IGNORECASE
    )

    # NOTE:
    # ---
    # [\"\']?(\$(?:(?P<escaped>\$|\d+)|({(?P<braced>.*?)(:(?P<braced_default>.*?))?})))[\"\']?
    #
    # NOTE: If you want to catch only string not number:
    #   ... `([a-zA-Z0-9_.\s'\"\[\]\(\)]+?)`
    __re_env_search: str = r"""
        [\"\']?                             # single or double quoted value
        (\$(?:                              # start with non-capturing group
            (?P<escaped>\$|\d+)             # escape $ or number like $1
            |
            (\{
                (?P<braced>([^{}]*?))       # value in braced {} not contain {}
                (:                          # : seperator for default
                    (?P<braced_default>[^{}]*?) # value default with sep :
                )?
            })
        ))
        [\"\']?                             # single or double quoted value
    """
    RE_ENV_SEARCH: Pattern = re.compile(
        __re_env_search, MULTILINE | UNICODE | IGNORECASE | VERBOSE
    )

    __re_env_value_quoted: str = r"""
        ^
            (?P<quoted>[\'\"`])     # single or double quoted value
            (?P<value>.*)\1
        $
    """
    RE_ENV_VALUE_QUOTED: Pattern = re.compile(
        __re_env_value_quoted,
        MULTILINE | UNICODE | VERBOSE,
    )

    # NOTE:
    __re_env_escape: str = r"\\([^$])"
    RE_ENV_ESCAPE: Pattern = re.compile(__re_env_escape, MULTILINE | UNICODE)



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.config as conf
import pytest
import yaml


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    tgt_path: Path = test_path / "conf_file_temp"
    tgt_path.mkdir(exist_ok=True)
    with open(tgt_path / "test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path
    shutil.rmtree(tgt_path)


def test_base_conf_read_file(target_path):
    bcf = conf.BaseConfFl(target_path)

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == bcf.load(name="conn_local_file")

    bcf.move(
        "test_01_conn.yaml",
        dest=target_path / "connections/test_01_conn.yaml",
    )

    bcf_temp = conf.BaseConfFl(target_path)
    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == bcf_temp.load(name="conn_local_file")

    assert (target_path / "connections/test_01_conn.yaml").exists()


def test_conf_read_file(target_path):
    cf = conf.ConfFl(target_path)
    cf.move(
        path="test_01_conn.yaml",
        dest=target_path / "connections/test_01_conn.yaml",
    )

    _stage_path: Path = target_path / "connections/test_01_conn_stage.json"

    cf.create(path=_stage_path)
    assert _stage_path.exists()
    cf.save_stage(path=_stage_path, data=cf.load("conn_local_file"))

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == cf.load_stage(path=_stage_path)

    cf.save_stage(
        path=_stage_path,
        data={"temp_additional": cf.load("conn_local_file")},
        merge=True,
    )

    cf.remove_stage(
        path=_stage_path,
        name="temp_additional",
    )

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == cf.load_stage(path=_stage_path)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
from typing import Callable

try:
    from .__regex import RegexConf
except ImportError:  # no cove
    from __regex import RegexConf


def add_newline(text: str, newline: str | None = None) -> str:
    """Add newline to a text value.

    :param text: A text value that want to add newline.
    :param newline: A newline value that want to use.

    :rtype: str
    :returns: A newline added text.
    """
    nl: str = newline or "\n"
    return f"{text}{nl}" if not text.endswith(nl) else text


def search_env_replace(
    contents: str,
    *,
    raise_if_default_not_exists: bool = False,
    default: str = "null",
    escape: str = "ESC",
    caller: Callable[[str], str] = (lambda x: x),
) -> str:
    """Prepare content data before parse to any file parsing object.

    :param contents: A string content that want to format with env vars
    :type contents: str
    :param raise_if_default_not_exists: A flag that will allow this function
        raise the error when default of env var does not set from contents.
    :type raise_if_default_not_exists: bool(=False)
    :param default: a default value.
    :type default: str(='null')
    :param escape: a escape value that use for initial replace when found escape
        char on searching.
    :type escape: str(='ESC')
    :param caller: a prepare function that will execute before replace env var.
    :type caller: Callable[[str], str]

    :rtype: str
    :returns: A prepared content data.

    Examples:

        >>> import os
        >>> os.environ["NAME"] = 'foo'
        >>> search_env_replace("Hello ${NAME}")
        'Hello foo'
    """
    shifting: int = 0
    replaces: dict = {}
    replaces_esc: dict = {}
    for content in RegexConf.RE_ENV_SEARCH.finditer(contents):
        search: str = content.group(1)
        if not (_escaped := content.group("escaped")):
            var: str = content.group("braced")
            _braced_default: str = content.group("braced_default")
            if not _braced_default and raise_if_default_not_exists:
                raise ValueError(
                    f"Could not find default value for {var} in the contents"
                )
            elif not var:
                raise ValueError(
                    f"Value {search!r} in the contents file has something "
                    f"wrong with regular expression"
                )
            replaces[search] = caller(
                os.environ.get(var, _braced_default) or default
            )
        elif "$" in _escaped:
            span = content.span()
            search = f"${{{escape}{_escaped}}}"
            contents = (
                contents[: (span[0] + shifting)]
                + search
                + contents[(span[1] + shifting) :]
            )
            shifting += len(search) - (span[1] - span[0])
            replaces_esc[search] = "$"
    for _replace in sorted(replaces, reverse=True):
        contents = contents.replace(_replace, replaces[_replace])
    for _replace in sorted(replaces_esc, reverse=True):
        contents = contents.replace(_replace, replaces_esc[_replace])
    return contents


def search_env(
    contents: str,
    *,
    keep_newline: bool = False,
    default: str | None = None,
) -> dict[str, str]:
    """Prepare content data from `.env` file before load to the OS environment
    variables.

    :param contents: A string content in the `.env` file
    :type contents: str
    :param keep_newline: A flag that filter out a newline
    :type keep_newline: bool(=False)
    :param default: A default value that use if it does not exists
    :type default: str | None(=None)

    :rtype: dict[str, str]
    :returns: A mapping of name and value of env variable

    Note:
        This function reference code from python-dotenv package. I will use this
    instead install this package. Because I want to enhance serialize step that
    fit with my package. (https://github.com/theskumar/python-dotenv)

    Examples:

        >>> search_env("Data='demo'\\nfoo=bar")
        {'Data': 'demo', 'foo': 'bar'}
        >>> search_env("Data='demo'\\n# foo=bar\\nhello=${Data}-2")
        {'Data': 'demo', 'hello': 'demo-2'}
    """
    _default: str = default or ""
    env: dict[str, str] = {}
    for content in RegexConf.RE_DOTENV.finditer(contents):
        name: str = content.group("name")

        # NOTE: Remove leading/trailing whitespace
        _value: str = (content.group("value") or "").strip()

        if not _value or _value in ("''", '""'):
            raise ValueError(
                f"Value {name!r} in `.env` file does not set value "
                f"of variable"
            )
        value: str = _value if keep_newline else "".join(_value.splitlines())
        quoted: str | None = None

        # Remove surrounding quotes
        if m2 := RegexConf.RE_ENV_VALUE_QUOTED.match(value):
            quoted: str = m2.group("quoted")
            value: str = m2.group("value")

        if quoted == "'":
            env[name] = value
            continue
        elif quoted == '"':
            # Unescape all chars except $ so variables
            # can be escaped properly
            value: str = RegexConf.RE_ENV_ESCAPE.sub(r"\1", value)

        # Substitute variables in a value
        env[name] = __search_var(value, env, default=_default)
    return env


def __search_var(
    value: str,
    env: dict[str, str],
    *,
    default: str | None = None,
) -> str:
    """Search variable on the string content.

    :param value: a string value that want to search env variable.
    :type value: str
    :param env: a pair of env values that keep in memory dict.
    :type env: dict[str, str]
    :param default: a default value if it does not found on env vars.
    :type default: str | None(=None)

    :rtype: str
    :returns: A searched value from env veriables.

    Examples:
        >>> __search_var("Test ${VAR}", {"VAR": "foo"})
        'Test foo'
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"})
        'Test '
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"}, default="bar")
        'Test bar'
        >>> import os
        >>> os.environ["VAR2"] = "baz"
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"}, default="bar")
        'Test baz'
    """
    _default: str = default or ""
    for sub_content in RegexConf.RE_DOTENV_VAR.findall(value):
        replace: str = "".join(sub_content[1:-1])
        if sub_content[0] != "\\":
            # Replace it with the value from the environment
            replace: str = env.get(
                sub_content[-1],
                os.environ.get(sub_content[-1], _default),
            )
        value: str = value.replace("".join(sub_content[:-1]), replace)
    return value



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from enum import Enum, IntEnum
from functools import total_ordering
from typing import TypeVar

T = TypeVar("T")


def enum_ordering(cls: T) -> T:
    """Add order property to Enum object."""

    def __lt__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self.value < other.value
        raise ValueError("Cannot compare different Enums")

    cls.__lt__ = __lt__
    return total_ordering(cls)


class StrEnum(str, Enum):
    """StrEnum where enum.auto() returns the field name.

    Reference:
        (https://docs.python.org/3.9/library/enum.html#using-automatic-values)
    """

    @staticmethod
    def _generate_next_value_(
        name: str,
        start: int,
        count: int,
        last_values: list,
    ) -> str:
        return name

    def __str__(self) -> str:
        return self.value


class Status(StrEnum):
    SUCCESS: str = "SUCCESS"
    APPROVED: str = "APPROVED"
    FAILED: str = "FAILED"
    WAITING: str = "WAITING"
    PROCESSING: str = "PROCESSING"
    TRIGGERED: str = "TRIGGERED"

    def in_process(self) -> bool:
        return self.value in ("WAITING", "PROCESSING")

    def is_done(self) -> bool:
        return self.value == "SUCCESS"

    def is_failed(self) -> bool:
        return self.value == "FAILED"


class Loading(StrEnum):
    FULL_DUMP = "F"
    DELTA = "D"
    MERGE = "D"
    TRANSACTION = "T"
    SCD_DELTA = "SCD_D"
    SCD_DUMP = "SCD_F"
    SCD_TRANS = "SCD_T"

    def is_delta(self) -> bool:
        return self.value == "D" or self.value == "SCD_D"

    def is_scd(self) -> bool:
        return self.value.startswith("SCD")


class DataLayer(IntEnum):
    RAW: int = 0
    STAGING: int = 1
    PERSISTED: int = 2
    CURATED: int = 3
    MART: int = 4
    REPORT: int = 5



import os
import shutil
from collections.abc import Generator
from pathlib import Path
from textwrap import dedent

import ddeutil.io.__base as bfl
import pytest


@pytest.fixture(scope="module")
def toml_path(test_path) -> Generator[Path, None, None]:
    this_path: Path = test_path / "toml"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_toml(toml_path):
    toml_file_path: Path = toml_path / "test_simple.toml"

    with open(toml_file_path, mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """
    [config]
    # Comment this line ...
    value = "foo"
    """
            ).strip()
        )

    assert {
        "config": {
            "value": "foo",
        },
    } == bfl.TomlFl(path=toml_file_path).read()


def test_toml_env(toml_path):
    toml_file_path: Path = toml_path / "test_env.toml"

    os.environ["TEST_TOML_ENV"] = "FOO"

    with open(toml_file_path, mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """
    [config]
    # Comment this line ...
    value = "foo is ${TEST_TOML_ENV}"
    """
            ).strip()
        )

    assert {
        "config": {
            "value": "foo is FOO",
        },
    } == bfl.TomlEnvFl(path=toml_file_path).read()



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.__base.files as fl
import ddeutil.io.__base.utils as utils
import pytest


@pytest.fixture(scope="module")
def openfile_path(test_path) -> Generator[Path, None, None]:
    this_path: Path = test_path / "open_file"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


@pytest.fixture(scope="module")
def encoding() -> str:
    return "utf-8"


def test_open_file_common(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_common_file.text",
        encoding=encoding,
    )
    with opf.open(mode="w") as f:
        f.write("Write data with common file in normal mode")

    with opf.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in normal mode" == rs

    with opf() as f:
        rs = f.read()

    assert "Write data with common file in normal mode" == rs


def test_open_file_common_append(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_common_file_append.text",
        encoding=encoding,
    )
    with opf.open(mode="w") as f:
        f.write(
            utils.add_newline(
                "Write data with common file append in normal mode",
            )
        )

    with opf.open(mode="a", newline="\n") as f:
        f.write("Write another line in the same file")

    with opf.open(mode="r") as f:
        rs = f.read()

    assert (
        "Write data with common file append in normal mode\n"
        "Write another line in the same file"
    ) == rs


def test_open_file_common_gzip(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_common_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with opf.open(mode="w") as f:
        f.write("Write data with common file in gzip mode")

    with opf.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in gzip mode" == rs


def test_open_file_common_xz(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_common_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with opf.open(mode="w") as f:
        f.write("Write data with common file in xz mode")

    with opf.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in xz mode" == rs


def test_open_file_common_bz2(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_common_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with opf.open(mode="w") as f:
        f.write("Write data with common file in bz2 mode")

    with opf.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in bz2 mode" == rs


def test_open_file_binary(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_binary_file.text",
        encoding=encoding,
    )
    with opf.open(mode="wb") as f:
        f.write(b"Write data with binary file in normal mode")

    with opf.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in normal mode" == rs


def test_open_file_binary_gzip(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_binary_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with opf.open(mode="wb") as f:
        f.write(b"Write data with binary file in gzip mode")

    with opf.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in gzip mode" == rs


def test_open_file_binary_xz(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_binary_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with opf.open(mode="wb") as f:
        f.write(b"Write data with binary file in xz mode")

    with opf.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in xz mode" == rs


def test_open_file_binary_bz2(openfile_path, encoding):
    opf = fl.Fl(
        path=openfile_path / "test_binary_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with opf.open(mode="wb") as f:
        f.write(b"Write data with binary file in bz2 mode")

    with opf.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in bz2 mode" == rs


@pytest.fixture(scope="module")
def openfile_mem_path(test_path) -> Generator[Path, None, None]:
    this_path: Path = test_path / "open_file_mem"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_open_file_mem_common(openfile_mem_path, encoding):
    opf = fl.Fl(
        path=openfile_mem_path / "test_common_mem_file.text",
        encoding=encoding,
    )
    with opf.mopen(mode="w") as f:
        f.write("Write data with common file in normal mode on memory")

    with opf.mopen(mode="r") as f:
        rs = f.read()

    assert b"Write data with common file in normal mode on memory" == rs


def test_open_file_mem_common_gzip(openfile_mem_path, encoding):
    opf = fl.Fl(
        path=openfile_mem_path / "test_common_mem_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with opf.mopen(mode="w") as f:
        f.write("Write data with common file in gzip mode on memory")

    with opf.mopen(mode="r") as f:
        rs = fl.compress_lib("gzip").decompress(f.read())

    assert b"Write data with common file in gzip mode on memory" == rs


def test_open_file_mem_common_xz(openfile_mem_path, encoding):
    opf = fl.Fl(
        path=openfile_mem_path / "test_common_mem_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with opf.mopen(mode="w") as f:
        f.write("Write data with common file in xz mode on memory")

    with opf.mopen(mode="r") as f:
        rs = fl.compress_lib("xz").decompress(f.read())

    assert b"Write data with common file in xz mode on memory" == rs


def test_open_file_mem_common_bz2(openfile_mem_path, encoding):
    opf = fl.Fl(
        path=openfile_mem_path / "test_common_mem_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with opf.mopen(mode="w") as f:
        f.write("Write data with common file in bz2 mode on memory")

    with opf.mopen(mode="r") as f:
        rs = fl.compress_lib("bz2").decompress(f.read())

    assert b"Write data with common file in bz2 mode on memory" == rs



from pathlib import Path

import ddeutil.io.param as md
import pytest
from ddeutil.io.__conf import UPDATE_KEY, VERSION_KEY
from ddeutil.io.exceptions import ConfigArgumentError


def test_model_path_default(test_path):
    p = md.PathData.model_validate(
        {
            "root": test_path,
        }
    )
    assert p.root == test_path


def test_model_path_data():
    p = md.PathData.model_validate(
        {
            "data": Path("."),
            "conf": Path("."),
        }
    )

    assert {
        "data": Path("."),
        "conf": Path("."),
        "root": Path("."),
    } == p.model_dump()


def test_model_path_data_with_root():
    p = md.PathData.model_validate({"root": "./src/"})
    assert {
        "data": Path("./src/data"),
        "conf": Path("./src/conf"),
        "root": Path("./src/"),
    } == p.model_dump()


def test_model_rule_data():
    assert {
        "timestamp": {},
        "version": None,
        "excluded": [],
        "compress": None,
    } == md.Rule.model_validate({}).model_dump()


def test_model_stage_data():
    stage = md.Stage.model_validate(
        {
            "alias": "persisted",
            "format": "{timestamp:%Y-%m-%d}{naming:%c}.json",
            "rules": {
                "timestamp": {"minutes": 15},
            },
        }
    )

    assert {
        "alias": "persisted",
        "format": "{timestamp:%Y-%m-%d}{naming:%c}.json",
        "rules": {
            "timestamp": {"minutes": 15},
            "version": None,
            "excluded": [],
            "compress": None,
        },
        "layer": 0,
    } == stage.model_dump()

    with pytest.raises(ConfigArgumentError):
        md.Stage.model_validate(
            {
                "alias": "persisted",
                "format": "timestamp.json",
                "rules": {
                    "timestamp": {"minutes": 15},
                },
            }
        )

    with pytest.raises(ConfigArgumentError):
        md.Stage.model_validate(
            {
                "alias": "persisted",
                "format": "{datetime:%Y%m%d}.json",
                "rules": {
                    "timestamp": {"minutes": 15},
                },
            }
        )


def test_model_params():
    params = md.Params.model_validate(
        {
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                "curated": {"format": "{domain:%s}_{naming:%s}.{compress:%-g}"},
            }
        }
    )
    assert {
        "stages": {
            "raw": {
                "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                "alias": "raw",
                "rules": {
                    "timestamp": {},
                    "version": None,
                    "excluded": [],
                    "compress": None,
                },
                "layer": 1,
            },
            "persisted": {
                "format": "{naming:%s}.{version:v%m.%n.%c}",
                "alias": "persisted",
                "rules": {
                    "timestamp": {},
                    "version": None,
                    "excluded": [],
                    "compress": None,
                },
                "layer": 2,
            },
            "curated": {
                "format": "{domain:%s}_{naming:%s}.{compress:%-g}",
                "alias": "curated",
                "rules": {
                    "timestamp": {},
                    "version": None,
                    "excluded": [],
                    "compress": None,
                },
                "layer": 3,
            },
        },
        "engine": {
            "values": {
                "dt_fmt": "%Y-%m-%d %H:%M:%S",
                "excluded": (VERSION_KEY, UPDATE_KEY),
            },
            "paths": {
                "conf": Path("conf"),
                "data": Path("data"),
                "root": Path("."),
            },
        },
    } == params.model_dump()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from typing import Literal

from ..__base import BaseUpdatableModel
from .col import Col


class BaseFl(BaseUpdatableModel):
    name: str
    encoding: str = "utf-8"


class CsvFl(BaseFl):
    """Csv File Model"""

    type: Literal["csv"] = "csv"
    header: bool = True
    feature: list[Col]
    sep: str = ","
    comment: str = "#"
    skip_rows: int = 0
    skip_footer: int = 0
    quote_char: str = '"'


class JsonFl(BaseFl):
    """Json File Model"""

    type: Literal["json"] = "json"
    nestest: int = -1


class ParqFl(BaseFl):
    """Parquet File Model"""

    type: Literal["parquet"] = "parquet"
    compress: str = "gzip"



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import logging
import os
from datetime import datetime
from typing import (
    Any,
    Optional,
    TypedDict,
)

from dateutil.relativedelta import relativedelta
from ddeutil.core import (
    concat,
    hash_all,
    merge_dict,
)
from ddeutil.core.__base import must_rsplit
from ddeutil.core.dtutils import get_date
from deepdiff import DeepDiff
from fmtutil import (
    ConstantType,
    Datetime,
    FormatterArgumentError,
    FormatterGroup,
    FormatterGroupType,
    Naming,
    VerPackage,
    Version,
    make_const,
    make_group,
)

from .__base import Fl, rm
from .__conf import UPDATE_KEY, VERSION_KEY
from .config import ConfFl
from .exceptions import ConfigArgumentError, ConfigNotFound
from .param import Params

METADATA: dict[str, Any] = {}


class StageFiles(TypedDict):
    parse: FormatterGroup
    file: str


CompressConst: ConstantType = make_const(
    name="CompressConst",
    formatter={
        "%g": "gzip",
        "%-g": "gz",
        "%b": "bz2",
        "%r": "rar",
        "%x": "xz",
        "%z": "zip",
    },
)

FileExtensionConst: ConstantType = make_const(
    name="FileExtensionConst",
    formatter={
        "%j": "json",
        "%y": "yaml",
        "%e": "env",
        "%t": "toml",
    },
)


class BaseRegister:
    """Base Register Object"""

    def __init__(
        self,
        name: str,
        *,
        domain: Optional[str] = None,
    ) -> None:
        self.name: str = name
        self.updt: datetime = get_date("datetime")
        self.domain: str = (
            domain.replace(os.sep, "/").strip("/").lower() if domain else ""
        )
        if any(sep in self.name for sep in (",", ".")):
            raise ConfigArgumentError(
                "name",
                "Config name should not contain `,` or `.` char",
            )

    @property
    def fullname(self) -> str:
        """Return a configuration fullname, which join `name` and `domain`
        together with domain partition string.
        """
        return f"{self.domain}:{self.name}" if self.domain else self.name

    @property
    def shortname(self) -> str:
        """Return a configuration shortname, which get first character of any
        split string by name partition string.
        """
        return concat(word[0] for word in self.name.split("_"))

    @property
    def fmt_group(self) -> FormatterGroupType:
        """Generate the formatter group that include constant formatters from
        ``self.name`` and ``self.domain``.
        """
        return make_group(
            {
                "naming": make_const(fmt=Naming, value=self.name),
                "domain": make_const(fmt=Naming, value=self.domain),
                "compress": CompressConst,
                "extension": FileExtensionConst,
                "version": Version,
                "timestamp": Datetime,
            }
        )


class Register(BaseRegister):
    """Register Object that contain configuration loading methods and metadata
    management. This object work with stage input argument, that set all
    properties in the `parameter.yaml` file.
    """

    @classmethod
    def reset(
        cls,
        name: str,
        params: Params,
    ) -> Register:
        """Reset all configuration data files that exists in any stage but
        does not do anything in the base stage. This method will use when the
        config name of data was changed and does not use the old name. If the
        name was changed and that config data does not reset,
        the configuration files of this data will exist in any moved stage.

        :param name: The fullname of configuration.
        :type name: str
        :param params:
        :type params: Params
        """
        for stage in params.stages:
            try:
                cls(name, stage=stage, params=params).remove()
            except ConfigNotFound:
                continue
        return cls(name, params=params)

    def __init__(
        self,
        name: str,
        stage: str | None = None,
        *,
        params: Params | None = None,
        loader: type[Fl] | None = None,
        loader_stg: type[Fl] | None = None,
    ):
        _domain, _name = must_rsplit(concat(name.split()), ":", maxsplit=1)
        super().__init__(name=_name, domain=_domain)
        if not params:
            raise NotImplementedError(
                "This register instance can not do any actions because config "
                "param does not set."
            )
        self.stage: str = stage or "base"
        self.loader: Optional[type[Fl]] = loader
        self.loader_stg: Optional[type[Fl]] = loader_stg
        self.params: Optional[Params] = params

        # Load latest version of data from data lake or data store of
        # configuration files
        self.__data: dict[str, Any] = self.pick(stage=self.stage)
        if not self.__data:
            raise ConfigNotFound(
                f"Config {self.name!r} "
                f"{f'in domain {self.domain!r} ' if self.domain else ' '}"
                f"does not exist in stage {self.stage!r}."
            )

        self.meta: dict[str, Any] = METADATA.get(self.fullname, {})

        # NOTE:
        #   Compare data from current stage and latest version in metadata.
        self.changed: int = self.compare_data(
            target=self.meta.get(self.stage, {})
        )

        # NOTE:
        #   Update metadata if the configuration data does not exist, or
        #   it has any changes.
        if self.changed == 99:
            logging.info(
                f"Configuration data with stage: {self.stage!r} does not "
                f"exists in metadata ..."
            )
            # TODO: Create metadata for caching value before compare data next
            #   time. (It can be table on database or sqlite file.
            # METADATA.update({"self.fullname": self.__data})
        elif self.changed > 0:
            logging.info(
                f"Should update metadata because diff level is {self.changed}."
            )

        # TODO: Remove this line when develop metadata feature in the next
        #   release.
        METADATA.pop(self.fullname, None)

    def __hash__(self) -> int:
        return hash(
            self.fullname
            + self.stage
            + f"{self.timestamp:{self.params.engine.values.dt_fmt}}"
        )

    def __str__(self) -> str:
        return f"({self.fullname}, {self.stage})"

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(name={self.fullname!r}"
            f"{f'stage={self.stage!r}' if self.stage != 'base' else ''})>"
        )

    def __eq__(self, other: Register) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.fullname == other.fullname
                and self.stage == other.stage
                and self.timestamp == other.timestamp
            )
        return NotImplemented

    def data(self, hashing: bool = False) -> dict[str, Any]:
        """Return the data with the configuration name."""
        _data = self.__data
        if not self.stage or (self.stage == "base"):
            _data = {
                k: v
                for k, v in (self.meta.get(self.stage, {}).items())
                if k in self.params.engine.values.excluded
            } | self.__data
        return (
            hash_all(_data, exclude=set(self.params.engine.values.excluded))
            if hashing
            else _data
        )

    @property
    def timestamp(self) -> datetime:
        """Return the current timestamp value of config data. If timestamp value
        does not exist. this property will return timestamp of initialize.

        :return: datetime
        """
        if self.changed > 0:
            return self.updt
        elif _dt := self.data().get(UPDATE_KEY):
            return datetime.strptime(
                _dt,
                self.params.engine.values.dt_fmt,
            )
        return self.updt

    def version(self, _next: bool = False) -> VerPackage:
        """Generate version value from the pick method. If version value does
        not exist from configuration data, this property will return the
        default, `v0.0.1`. If the initialization process tracking some change
        from configuration data between metadata and the latest data in the
        stage, the _next will be generated.

        :return: VerPackage
        """
        _vs = VerPackage.parse(self.data().get(VERSION_KEY, "v0.0.1"))
        if not _next or self.changed == 0:
            return _vs
        elif self.changed >= 3:
            return _vs.bump_major()
        elif self.changed == 2:
            return _vs.bump_minor()
        return _vs.bump_patch()

    def fmt(self, update: Optional[dict[str, Any]] = None) -> FormatterGroup:
        return self.fmt_group(
            {
                "timestamp": self.timestamp,
                "version": self.version(),
                **(update or {}),
            }
        )

    def compare_data(
        self,
        target: dict[Any, Any],
    ) -> int:
        """Return difference column from dictionary comparison method which use
        the `deepdiff` library.

        :param target: dict : The target dictionary for compare with current
                configuration data.
        """
        if not target:
            return 99

        results = DeepDiff(
            self.data(hashing=True),
            target,
            ignore_order=True,
            exclude_paths={
                f"root[{key!r}]" for key in self.params.engine.values.excluded
            },
        )
        if any(
            _ in results
            for _ in (
                "dictionary_item_added",
                "dictionary_item_removed",
                "iterable_item_added",
                "iterable_item_removed",
            )
        ):
            return 2
        elif any(
            _ in results
            for _ in (
                "values_changed",
                "type_changes",
            )
        ):
            return 1
        return 0

    def __stage_files(
        self,
        stage: str,
        loading: ConfFl,
    ) -> dict[int, StageFiles]:
        """Return mapping of StageFiles data."""
        results: dict[int, StageFiles] = {}
        for index, file in enumerate(
            (_f.name for _f in loading.files()),
            start=1,
        ):
            try:
                results[index]: dict = {
                    "parse": self.fmt_group.parse(
                        value=file,
                        fmt=rf"{self.params.get_stage(stage).format}\.json",
                    ),
                    "file": file,
                }
            except FormatterArgumentError:
                continue
        return results

    def pick(
        self,
        stage: Optional[str] = None,
        *,
        order: Optional[int] = 1,
        reverse: bool = False,
    ) -> dict[str, Any]:
        if (stage is None) or (stage == "base"):
            return ConfFl(
                path=(self.params.engine.paths.conf / self.domain),
                open_file=self.loader,
                open_file_stg=self.loader_stg,
            ).load(name=self.name, order=order)

        loading = ConfFl(
            path=self.params.engine.paths.data / stage,
            compress=self.params.get_stage(stage).rules.compress,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )

        if results := self.__stage_files(stage, loading):
            max_data: list = sorted(
                results.items(),
                key=lambda x: (x[1]["parse"],),
                reverse=reverse,
            )
            return loading.load_stage(
                path=(loading.path / max_data[-order][1]["file"])
            )
        return {}

    def move(
        self,
        stage: str,
        *,
        force: bool = False,
        retention: bool = True,
    ) -> Register:
        """Move file to the target stage."""
        loading: ConfFl = ConfFl(
            path=self.params.engine.paths.data / stage,
            compress=self.params.get_stage(stage).rules.compress,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )
        if (
            self.compare_data(
                hash_all(
                    self.pick(stage=stage),
                    exclude=set(self.params.engine.values.excluded),
                )
            )
            > 0
            or force
        ):
            _filename: str = self.fmt().format(
                f"{self.params.get_stage(name=stage).format}.json",
            )
            if (loading.path / _filename).exists():
                # TODO: generate serial number if file exists
                logging.warning(
                    f"File {_filename!r} already exists in {stage!r} stage."
                )
            _dt_fmt: str = self.params.engine.values.dt_fmt
            loading.save_stage(
                path=(loading.path / _filename),
                data=merge_dict(
                    self.data(),
                    {
                        UPDATE_KEY: f"{self.timestamp:{_dt_fmt}}",
                        VERSION_KEY: f"v{str(self.version())}",
                    },
                ),
            )
            # NOTE:
            #   Retention process after move data to the stage successful
            if retention:
                self.purge(stage=stage)
        else:
            logging.warning(
                f"Config {self.name!r} cannot move {self.stage!r} -> "
                f"{stage!r} cause the data does not has any change or "
                f"force moving flag does not set."
            )
        return self.switch(stage=stage)

    def switch(self, stage: str) -> Register:
        """Switch instance from old stage to new stage with input argument."""
        return self.__class__(
            name=self.fullname,
            stage=stage,
            params=self.params,
        )

    def purge(self, stage: Optional[str] = None) -> None:
        """Purge configuration files that match with any rules in the stage
        setting.
        """
        _stage: str = stage or self.stage
        if not (_rules := self.params.get_stage(_stage).rules):
            return
        loading: ConfFl = ConfFl(
            path=self.params.engine.paths.data / stage,
            compress=_rules.compress,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )
        rs: dict[int, StageFiles] = self.__stage_files(_stage, loading)
        max_file: FormatterGroup = max(
            rs.items(),
            key=lambda x: (x[1]["parse"],),
        )[1]["parse"]

        upper_bound: Optional[FormatterGroup] = None
        if _rtt_ts := _rules.timestamp:
            upper_bound = max_file.adjust(
                {"timestamp": relativedelta(**_rtt_ts)}
            )

        if upper_bound is not None:
            for _, data in filter(
                lambda x: x[1]["parse"] < upper_bound,
                rs.items(),
            ):
                _file: str = data["file"]
                rm(loading.path / _file)

    def deploy(self, stop: Optional[str] = None) -> Register:
        """Deploy data that move from base to final stage.

        :param stop: A stage name for stop when move config from base stage
            to final stage.
        :type stop: str
        """
        _base: Register = self
        _stop: str = stop or self.params.stage_final
        assert (
            _stop in self.params.stages
        ), "a `stop` argument should exists in stages data on Param config."
        for stage in self.params.stages:
            _base: Register = _base.move(stage)
            if _stop and (stage == _stop):
                break
        return _base

    def remove(self, stage: Optional[str] = None) -> None:
        """Remove config file from the stage storage.

        :param stage: a stage value that want to remove.
        :type stage: Optional[str]
        """
        _stage: str = stage or self.stage
        assert (
            _stage != "base"
        ), "The remove method can not process on the 'base' stage."
        loading: ConfFl = ConfFl(
            path=self.params.engine.paths.data / _stage,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )

        # Remove all files from the stage.
        data: StageFiles
        for _, data in self.__stage_files(_stage, loading).items():
            _file: str = data["file"]
            rm(loading.path / _file)


class FullRegister(Register):
    """Full register that implement archiving step on base Register."""

    def purge(self, stage: Optional[str] = None) -> None:
        """Purge configuration files that match with any rules in the stage
        setting.
        """
        _stage: str = stage or self.stage
        if not (_rules := self.params.get_stage(_stage).rules):
            return
        loading: ConfFl = ConfFl(
            path=self.params.engine.paths.data / stage,
            compress=_rules.compress,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )
        rs: dict[int, StageFiles] = self.__stage_files(_stage, loading)
        max_file: FormatterGroup = max(
            rs.items(),
            key=lambda x: (x[1]["parse"],),
        )[1]["parse"]

        upper_bound: Optional[FormatterGroup] = None
        if _rtt_ts := _rules.timestamp:
            upper_bound = max_file.adjust(
                {"timestamp": relativedelta(**_rtt_ts)}
            )

        if upper_bound is not None:
            for _, data in filter(
                lambda x: x[1]["parse"] < upper_bound,
                rs.items(),
            ):
                _file: str = data["file"]
                # NOTE: Archive step
                _ac_path: str = (
                    f"{stage.lower()}_{self.updt:%Y%m%d%H%M%S}_{_file}"
                )
                loading.move(
                    _file,
                    dest=self.params.engine.paths.data / ".archive" / _ac_path,
                )
                rm(loading.path / _file)

    def remove(self, stage: Optional[str] = None) -> None:
        """Remove config file from the stage storage.

        :param stage: a stage value that want to remove.
        :type stage: Optional[str]
        """
        _stage: str = stage or self.stage
        assert (
            _stage != "base"
        ), "The remove method can not process on the 'base' stage."
        loading: ConfFl = ConfFl(
            path=self.params.engine.paths.data / _stage,
            open_file=self.loader,
            open_file_stg=self.loader_stg,
        )

        # NOTE: Remove all files from the stage.
        data: StageFiles
        for _, data in self.__stage_files(_stage, loading).items():
            _file: str = data["file"]
            # NOTE: Archive step
            _ac_path: str = f"{_stage.lower()}_{self.updt:%Y%m%d%H%M%S}_{_file}"
            loading.move(
                _file,
                dest=self.params.engine.paths / ".archive" / _ac_path,
            )
            rm(loading.path / _file)


__all__ = (
    "Register",
    "FullRegister",
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import re
from typing import (
    Annotated,
    Any,
    Optional,
    Union,
)

from ddeutil.core import onlyone
from pydantic import Field
from pydantic.functional_validators import (
    field_validator,
    model_validator,
)

from ..__base import BaseUpdatableModel
from ..const import Ref
from ..dtype import Dtype
from ..utils import (
    catch_str,
    extract_dtype,
    split_dtype,
)

DTYPE_KEYS: tuple[str, ...] = ("dtype", "datatype", "type")


class BaseCol(BaseUpdatableModel):
    """Base Column Model"""

    name: Annotated[
        str,
        Field(
            description="Name of Column",
            alias="Name",
        ),
    ]
    desc: Annotated[
        Optional[str], Field(description="Description of Column")
    ] = None
    dtype: Annotated[
        Dtype,
        Field(
            union_mode="left_to_right",
            description="Data Type of Column",
            alias="DataType",
        ),
    ]

    @field_validator("name")
    def __after_name(cls, value: str) -> str:
        """Prepare string value of name"""
        return "".join(value.strip().split())

    @field_validator("dtype", mode="before")
    def __before_str2dtype(cls, value: Union[str, dict, Dtype]) -> Dtype:
        """Prepare string value of dtype"""
        if isinstance(value, str):
            return extract_dtype(value)
        return value


class Col(BaseCol):
    """Column model"""

    nullable: Annotated[
        bool,
        Field(
            description="Nullable flag",
            alias="Nullable",
        ),
    ] = True
    unique: Annotated[
        bool,
        Field(
            description="Unique flag",
            alias="Unique",
        ),
    ] = False
    default: Annotated[
        Union[int, str, None],
        Field(
            description="Default value for this column",
            alias="Default",
        ),
    ] = None
    check: Annotated[
        Optional[str],
        Field(
            description="Check Statement",
            alias="Check",
        ),
    ] = None

    # NOTE:
    #   Special value that effect to parent model that include this model
    pk: Annotated[
        bool,
        Field(
            description="Primary key flag which can not contain null value",
            alias="PrimaryKey",
            repr=False,
        ),
    ] = False
    fk: Annotated[
        Union[Ref, dict],
        Field(
            default_factory=dict,
            description="Foreign key reference",
            alias="ForeignKey",
            repr=False,
        ),
    ]

    @classmethod
    def extract_column_from_dtype(
        cls,
        value: str,
    ) -> dict[str, Union[str, bool]]:
        """Extract a mapping column model from a string datatype."""
        values: dict[str, Any] = {"nullable": False}
        _dtype, _nullable = split_dtype(value)

        # Remove unique value from datatype
        _dtype, values["unique"] = catch_str(_dtype, key="unique")

        # Remove primary key value from datatype
        _dtype, values["pk"] = catch_str(_dtype, key="primary key")

        # Rename serial value to int from datatype
        _dtype, serial_flag = catch_str(_dtype, key="serial", replace="integer")

        if "check" in _dtype:
            if m := re.search(
                r"check\s?\((?P<check>[^()]*(?:\(.*\))*[^()]*)\)",
                _dtype,
            ):
                _dtype, values["check"] = catch_str(
                    _dtype, m.group(), flag=False
                )
            else:
                raise ValueError(
                    "datatype with type string does not support for "
                    "this format of check"
                )

        if re.search("default", _dtype):
            _dtype, _default = _dtype.split("default", maxsplit=1)
            values["dtype"] = _dtype.strip()
            values["default"] = _default.strip()
        else:
            values["dtype"] = _dtype

            if serial_flag:
                _nullable: str = "not null"
                # TODO:
                #   Change this default value for serial to dynamic value.
                values["default"] = "nextval('tablename_colname_seq')"

        if not values["pk"]:
            values["nullable"] = not re.search("not null", _nullable)
        return values

    @model_validator(mode="before")
    def __before_dtype(cls, values):
        """Prepare datatype value that parsing to this model with different
        types, string or dict type.

        This filter will prepare datatype value from the format,
            {DATATYPE} {UNIQUE} {NULLABLE} {DEFAULT}
            {PRIMARY KEY|FOREIGN KEY} {CHECK}

        Examples:
        - varchar( 100 ) not null default 'O' check( <name> <> 'test' )
        - serial not null primary key
        """
        if not (dtype_key := onlyone(values, DTYPE_KEYS, default=False)):
            raise ValueError("dtype key does not contain in values")

        pre_dtype: Any = values.pop(dtype_key)
        values_update: dict[str, Any] = {}
        if isinstance(pre_dtype, str):
            values_update = cls.extract_column_from_dtype(pre_dtype)
        else:
            values["dtype"] = pre_dtype

        return values_update | values

    @model_validator(mode="after")
    def __validate_logic(self):
        """Validate and check logic of values"""
        nullable: bool = self.nullable
        default: Union[int, str, None] = self.default

        # RULE: primary key and nullable does not True together
        if self.pk and nullable:
            # Raise: ValueError("`pk` and `nullable` can not be True together")
            self.nullable = False
        if (default is not None) and nullable:
            raise ValueError("`nullable` can not be True if `default` was set")
        return self



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import (
    Annotated,
    Literal,
    Union,
)

from pydantic import BaseModel, Field
from pydantic.functional_validators import field_validator


class BaseType(BaseModel):
    """Base Type"""

    type: str = "base"

    def __str__(self) -> str:
        return self.type


class StringType(BaseType):
    """String Type

    Note:
        This type will be the base of any string type that able to fixed length
    or not.
    """

    type: Literal["string", "str"] = "string"
    max_length: Annotated[int, Field(ge=-1)] = -1

    def __str__(self) -> str:
        _length: str = f"( {self.max_length} )" if self.max_length > -1 else ""
        return f"{self.type}{_length}"


class CharType(StringType):
    """Charactor Type

    Note: fixed-length strings
    """

    type: Literal["char"] = "char"


class VarcharType(StringType):
    """Variable Charactor Type

    Note: variable-length strings with limit
    """

    type: Literal["varchar"] = "varchar"


class TextType(StringType):
    """Text Type

    Note: variable unlimited length strings
    """

    type: Literal["text"] = "text"


class IntegerType(BaseType):
    """Integer Type

    Note: Storage size, 4 bytes, -2147483648 to +2147483647
    """

    type: Literal["integer", "int"] = "integer"

    @field_validator("type", mode="after")
    def prepare_for_short_name(
        cls,
        value: Literal["integer", "int"],
    ) -> Literal["integer"]:
        return "integer" if value == "int" else value


class SmallIntType(IntegerType):
    """Small Range Integer

    Note: Storage size, 2 bytes, -32768 to +32767
    """

    type: Literal["smallint"] = "smallint"


class BigIntType(IntegerType):
    """Big Range Integer

    Note: Storage size, 8 bytes, -9223372036854775808 to +9223372036854775807
    """

    type: Literal["bigint"] = "bigint"


class ShortType(SmallIntType):

    type: Literal["short"] = "short"


class LongType(BigIntType):

    type: Literal["long"] = "long"


class SerialType(IntegerType):
    """Serial Type"""

    type: Literal["serial"] = "serial"


class NumericType(BaseType):
    """Numeric Type"""

    type: Literal["numeric"] = "numeric"
    precision: Annotated[int, Field(ge=-1)] = -1
    scale: Annotated[int, Field(ge=-1)] = -1

    def __str__(self) -> str:
        if self.precision > -1:
            _scale: str = f", {self.scale}" if self.scale > -1 else ""
            return f"{self.type}( {self.precision}{_scale} )"
        return self.type


class DecimalType(NumericType):
    """Decimal Type"""

    type: Literal["decimal"] = "decimal"


class FloatType(BaseType):
    type: Literal["float"] = "float"


class RealType(BaseType):
    type: Literal["real"] = "real"


class DoublePrecisionType(BaseType): ...


class TimestampType(BaseType):
    """Timestamp Type"""

    type: Literal["timestamp"] = "timestamp"
    precision: Annotated[int, Field(ge=-1, le=6)] = -1
    timezone: Annotated[bool, Field(description="Timezone flag")] = False


class TimeType(BaseType):

    type: Literal["time"] = "time"


class DateType(BaseType):

    type: Literal["date"] = "date"


class DateTimeType(BaseType):

    type: Literal["datetime"] = "datetime"


Dtype = Union[
    StringType,
    CharType,
    VarcharType,
    TextType,
    NumericType,
    DecimalType,
    TimestampType,
    IntegerType,
    BigIntType,
    SmallIntType,
    BaseType,
]



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""
Define Errors Object for IO package
"""
from __future__ import annotations


class BaseError(Exception):
    """Base Error Object that use for catch any errors statement of
    all step in this src
    """


class IOBaseError(BaseError):
    """Core Base Error object"""


class ConfigNotFound(IOBaseError):
    """Error raise for a method not found the config file or data."""


class ConfigArgumentError(IOBaseError):
    """Error raise for a wrong configuration argument."""

    def __init__(self, argument: str | tuple, message: str) -> None:
        """Main Initialization that merge the argument and message input values
        with specific content message together like

            `__class__` with `argument`, `message`

        :param argument: Union[str, tuple]
        :param message: str
        """
        if isinstance(argument, tuple):
            _last_arg: str = str(argument[-1])
            _argument: str = (
                (
                    ", ".join([f"{_!r}" for _ in argument[:-1]])
                    + f", and {_last_arg!r}"
                )
                if len(argument) > 1
                else f"{_last_arg!r}"
            )
        else:
            _argument: str = f"{argument!r}"
        _message: str = f"with {_argument}, {message}"
        super().__init__(_message)



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.register as rgt
import pytest
import yaml
from ddeutil.io.param import Params


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    tgt_path: Path = test_path / "register_temp"
    tgt_path.mkdir(exist_ok=True)
    (tgt_path / "conf/demo").mkdir(parents=True)
    with open(tgt_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path
    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def params(target_path, root_path) -> Params:
    return Params.model_validate(
        {
            "engine": {
                "paths": {
                    "conf": target_path / "conf",
                    "data": root_path / "data",
                    "archive": root_path / "/data/.archive",
                },
                "flags": {"auto_update": True},
            },
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
            },
        }
    )


def test_register_init(params: Params):
    register = rgt.Register(name="demo:conn_local_file", params=params)

    assert "base" == register.stage
    assert {
        "alias": "conn_local_file",
        "type": "connection.LocalFileStorage",
        "endpoint": "file:///null/tests/examples/dummy",
    } == register.data()
    assert {
        "alias": "62d877a16819c672578d7bded7f5903c",
        "type": "cece9f1b3f4791a04ec3d695cb5ba1a9",
        "endpoint": "0d1db48bb2425db014fc66734508098f",
    } == register.data(hashing=True)

    print("\nChange compare from metadata:", register.changed)
    assert register.changed == 99

    rsg_raw = register.move(stage="raw")

    assert register.stage != rsg_raw.stage
    assert (
        "62d877a16819c672578d7bded7f5903c"
        == rsg_raw.data(hashing=True)["alias"]
    )

    rsg_persisted = rsg_raw.move(stage="persisted")
    assert rsg_raw.stage != rsg_persisted.stage
    assert (
        "62d877a16819c672578d7bded7f5903c"
        == rsg_persisted.data(hashing=True)["alias"]
    )
    rgt.Register.reset(name="demo:conn_local_file", params=params)


def test_register_without_params():
    try:
        rgt.Register(name="demo:conn_local_file")
    except NotImplementedError as err:
        assert (
            "This register instance can not do any actions because config "
            "param does not set."
        ) == str(err)



from .conn import (
    Conn,
    DbConn,
    FlConn,
)
from .const import (
    Fk,
    Pk,
    Ref,
)
from .dtype import (
    BaseType,
    BigIntType,
    CharType,
    DecimalType,
    Dtype,
    IntegerType,
    NumericType,
    SmallIntType,
    StringType,
    TextType,
    TimestampType,
    VarcharType,
)



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.__base as base
import pytest


@pytest.fixture(scope="module")
def make_empty_path(test_path: Path) -> Generator[Path, None, None]:
    path_search = test_path / "test_empty_path_search"
    path_search.mkdir(exist_ok=True)

    yield path_search

    shutil.rmtree(path_search)


@pytest.fixture(scope="module")
def make_path(test_path: Path) -> Generator[Path, None, None]:
    path_search = test_path / "test_path_search"
    path_search.mkdir(exist_ok=True)

    (path_search / "dir01").mkdir(exist_ok=True)
    base.touch(path_search / "dir01" / "01_01_test.text")
    base.touch(path_search / "dir01" / "01_02_test.text")
    (path_search / "dir02").mkdir(exist_ok=True)
    base.touch(path_search / "dir02" / "02_01_test.text")

    yield path_search

    shutil.rmtree(path_search)


def test_base_path_search_empty(make_empty_path):
    ps = base.PathSearch(make_empty_path)
    assert [] == ps.files
    assert 1 == ps.level


def test_base_path_search_raise(make_empty_path):
    with pytest.raises(FileNotFoundError):
        base.PathSearch(make_empty_path / "demo")


def test_base_path_search(make_path):
    ps = base.PathSearch(make_path)
    assert {
        make_path / "dir01/01_01_test.text",
        make_path / "dir01/01_02_test.text",
        make_path / "dir02/02_01_test.text",
    } == set(ps.files)

    ps = base.PathSearch(make_path, exclude=["dir02"])
    assert {
        make_path / "dir01/01_01_test.text",
        make_path / "dir01/01_02_test.text",
    } == set(ps.files)



import shutil
from collections.abc import Generator
from pathlib import Path

import ddeutil.io.config as conf
import pytest


@pytest.fixture(scope="module")
def sqlite_path(test_path) -> Generator[Path, None, None]:
    sqlite_path: Path = test_path / "conf_sqlite_temp"

    yield sqlite_path

    if sqlite_path.exists():
        shutil.rmtree(sqlite_path)


def test_base_conf_read_file(sqlite_path: Path):
    _schemas: dict[str, str] = {
        "name": "varchar(256) primary key",
        "shortname": "varchar(64) not null",
        "fullname": "varchar(256) not null",
        "data": "json not null",
        "updt": "datetime not null",
        "rtdt": "datetime not null",
        "author": "varchar(512) not null",
    }

    bc_sql = conf.ConfSQLite(sqlite_path)
    bc_sql.create(table="demo.db/temp_table", schemas=_schemas)

    assert (sqlite_path / "demo.db").exists()

    _data = {
        "conn_local_data_landing": {
            "name": "conn_local_data_landing",
            "shortname": "cldl",
            "fullname": "conn_local_data_landing",
            "data": {"first_row": {"key": "value"}},
            "updt": "2023-01-01 00:00:00",
            "rtdt": "2023-01-01 00:00:00",
            "author": "unknown",
        },
    }

    bc_sql.save_stage(table="demo.db/temp_table", data=_data)
    assert _data == bc_sql.load_stage(table="demo.db/temp_table")



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import fnmatch
import os
import shutil
from pathlib import Path
from typing import Optional

from .__regex import RegexConf
from .files import (
    CsvFl,
    CsvPipeFl,
    EnvFl,
    Fl,
    JsonEnvFl,
    JsonFl,
    MarshalFl,
    MsgpackFl,
    PickleFl,
    TomlEnvFl,
    TomlFl,
    YamlEnvFl,
    YamlFl,
)
from .utils import (
    add_newline,
    search_env_replace,
)


def rm(path: str, is_dir: bool = False) -> None:  # no cove
    """Remove file or dir from a input path."""
    if os.path.isfile(path) or os.path.islink(path):
        os.remove(path)
    elif os.path.isdir(path) and is_dir:
        shutil.rmtree(path)
    else:
        raise ValueError(
            f"file {path!r} is not a file{' or dir' if is_dir else ''}."
        )


def touch(filename: str, times=None) -> None:  # no cove
    """Touch file"""
    file_handle = open(filename, mode="a")
    try:
        os.utime(filename, times)
    finally:
        file_handle.close()


class PathSearch:
    """Path Search object"""

    def __init__(
        self,
        root: Path,
        *,
        exclude: Optional[list] = None,
        max_level: int = -1,
        length: int = 4,
        icon: int = 1,
    ):
        self.root: Path = root
        self.exclude: list = exclude or []
        self.max_level: int = max_level
        self.length: int = length
        self.real_level: int = 0

        # Declare icon arguments
        self._icon_last: str = self.icons()[icon]["last"]
        self._icon_next: str = self.icons()[icon]["next"]
        self._icon: str = self.icons()[icon]["normal"]
        self._icon_length: int = len(self._icon)

        assert (
            self._icon_length + 1
        ) < self.length, "a `length` argument must gather than length of icon."

        self.output_buf: list = [f"[{self.root.stem}]"]
        self.files: list[Path] = []
        self.__recurse(self.root, list(self.root.iterdir()), "", 0)

    @property
    def level(self) -> int:
        """Return level of sub path from the root path."""
        return self.real_level + 1 if self.max_level == -1 else self.max_level

    def __recurse(
        self,
        path: Path,
        file_list: list[Path],
        prefix: str,
        level: int,
    ):
        """Path recursive method for generate buffer of tree and files."""
        if not file_list or (self.max_level != -1 and self.max_level <= level):
            return

        self.real_level: int = max(level, self.real_level)
        file_list.sort(key=lambda f: (path / f).is_file())
        for idx, sub_path in enumerate(file_list):
            if any(exc == sub_path.name for exc in self.exclude):
                continue

            full_path: Path = path / sub_path
            idc: str = self.__switch_icon(idx, len(file_list))
            if full_path.is_dir():
                self.output_buf.append(f"{prefix}{idc}[{sub_path}]")
                tmp_prefix: str = (
                    (
                        f"{prefix}{self._icon}"
                        f'{" " * (self.length - self._icon_length)}'
                    )
                    if len(file_list) > 1 and idx != len(file_list) - 1
                    else f'{prefix}{" " * self.length}'
                )
                self.__recurse(
                    full_path, list(full_path.iterdir()), tmp_prefix, level + 1
                )
            elif full_path.is_file():  # no cove
                self.output_buf.append(f"{prefix}{idc}{sub_path}")
                self.files.append(full_path)

    def pick(self, filename: str) -> list[Path]:
        """Return filename with match with input argument."""
        return list(
            filter(
                lambda file: fnmatch.fnmatch(file, f"*/{filename}"),
                self.files,
            )
        )

    def tree(self, newline: Optional[str] = None) -> str:  # no cove
        """Return path tree of root path."""
        return (newline or "\n").join(self.output_buf)

    def __switch_icon(self, number_now: int, number_all: int):
        return (
            self._icon_last
            if number_now == (number_all - 1)
            else self._icon_next
        )

    @staticmethod
    def icons() -> dict[int, dict[str, str]]:
        return {
            1: {"normal": "│", "next": "├─", "last": "└─"},
            2: {"normal": "┃", "next": "┣━", "last": "┗━"},
            3: {"normal": "│", "next": "├─", "last": "╰─"},
        }



import shutil
from collections.abc import Generator
from pathlib import Path
from textwrap import dedent

import ddeutil.io.__base as bfl
import pytest


@pytest.fixture(scope="module")
def env_path(test_path) -> Generator[Path, None, None]:
    this_path: Path = test_path / "env"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_env(env_path):
    env_path: Path = env_path / ".env"

    with open(env_path, mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """
    TEST=This is common value test
    # Comment this line ...
    COMMENT_TEST='This is common value test'  # This is inline comment
    QUOTE='single quote'
    DOUBLE_QUOTE="double quote"
    PASSING=${DOUBLE_QUOTE}
    UN_PASSING='${DOUBLE_QUOTE}'
    """
            ).strip()
        )

    assert {
        "TEST": "This is common value test",
        "COMMENT_TEST": "This is common value test",
        "QUOTE": "single quote",
        "DOUBLE_QUOTE": "double quote",
        "PASSING": "double quote",
        "UN_PASSING": "${DOUBLE_QUOTE}",
    } == bfl.EnvFl(path=env_path).read(update=False)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""
This is then main function for open any files in local or remote space
with the best python libraries and the best practice such as build-in
``io.open``, ``mmap.mmap``, etc.

TODO: Add more compress type such as
    - h5,hdf5(h5py)
    - fits(astropy)
    - rar(...)
"""
from __future__ import annotations

import abc
import csv
import io
import json
import logging
import marshal
import mmap
import os
import pickle
import sys
import zipfile
from contextlib import contextmanager
from pathlib import Path
from tarfile import TarFile
from typing import (
    IO,
    Any,
    AnyStr,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Protocol,
    Union,
    get_args,
)

import msgpack
import toml
import yaml
from ddeutil.core import must_split

try:
    from yaml import CSafeLoader as SafeLoader
    from yaml import CUnsafeLoader as UnsafeLoader
except ImportError:  # no cove
    from yaml import SafeLoader, UnsafeLoader

from .utils import search_env, search_env_replace

FileCompressType = Literal["gzip", "gz", "xz", "bz2"]
DirCompressType = Literal["zip", "rar", "tar", "h5", "hdf5", "fits"]


def compress_lib(compress: str) -> CompressProtocol:
    """Return Compress module that use to unpack data from the compressed file.

    Note:
        Now, it support for "gzip", "gz", "xz", "bz2"]
    """
    if not compress:
        return io
    elif compress in ("gzip", "gz"):
        import gzip

        return gzip
    elif compress in ("bz2",):
        import bz2

        return bz2
    elif compress in ("xz",):
        import lzma as xz

        return xz
    raise NotImplementedError(f"Compress {compress} does not implement yet")


class CompressProtocol(Protocol):  # no cove
    def decompress(self, *args, **kwargs) -> AnyStr: ...

    def open(self, *args, **kwargs) -> IO: ...


class FlAbc(abc.ABC):  # no cove
    @abc.abstractmethod
    def read(self, *args, **kwargs): ...

    @abc.abstractmethod
    def write(self, *args, **kwargs): ...


class Fl(FlAbc):
    """Open File Object that use to open any simple or compression file from
    local file system.

    Examples:
        >>> with Fl('./dwh/conf/params.gz.txt', compress='gzip').open() as f:
        ...     data = f.readline()
    """

    def __init__(
        self,
        path: Union[str, Path],
        *,
        encoding: Optional[str] = None,
        compress: Optional[FileCompressType] = None,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        self.encoding: str = encoding or "utf-8"
        self.compress: Optional[FileCompressType] = compress

        # NOTE: Action anything after set up attributes.
        self.after_set_attrs()

    def after_set_attrs(self) -> None: ...

    def __call__(self, *args, **kwargs) -> IO:
        return self.open(*args, **kwargs)

    @property
    def decompress(self) -> Callable[[...], AnyStr]:
        if self.compress and self.compress in get_args(FileCompressType):
            return compress_lib(self.compress).decompress
        raise NotImplementedError(
            "Does not implement decompress method for None compress value."
        )

    def convert_mode(self, mode: str | None = None) -> dict[str, str]:
        """Convert mode before passing to the main standard lib.

        :param mode: a reading or writing mode for the open method.

        :rtype: dict[str, str]
        :returns: A mapping of mode and other input parameters for standard
            libs.
        """
        if not mode:
            return {"mode": "r"}
        byte_mode: bool = "b" in mode
        if self.compress is None:
            _mode: dict[str, str] = {"mode": mode}
            return _mode if byte_mode else {"encoding": self.encoding, **_mode}
        elif not byte_mode and self.compress in ("gzip", "gz", "xz", "bz2"):
            # NOTE:
            #   Add `t` in open file mode for force with text mode.
            return {"mode": f"{mode}t", "encoding": self.encoding}
        elif byte_mode and self.compress in ("gzip", "gz", "xz", "bz2"):
            return {"mode": mode}
        return {"mode": mode}

    def open(self, *, mode: Optional[str] = None, **kwargs) -> IO:
        return compress_lib(self.compress).open(
            self.path,
            **(self.convert_mode(mode) | kwargs),
        )

    @contextmanager
    def mopen(self, *, mode: Optional[str] = None) -> IO:
        mode: str = mode or "r"
        file: IO = self.open(mode=mode)
        _access = mmap.ACCESS_READ if ("r" in mode) else mmap.ACCESS_WRITE
        try:
            yield mmap.mmap(file.fileno(), length=0, access=_access)
        except ValueError as err:
            if str(err) != "cannot mmap an empty file":
                raise err
            yield file
            logging.error("Does not open file with memory mode")
        finally:
            file.close()

    def read(self, *args, **kwargs):
        raise NotImplementedError

    def write(self, *args, **kwargs):
        raise NotImplementedError


class OpenDirProtocol(Protocol):

    def write(self, name, arcname): ...

    def safe_extract(self, path, members): ...


class CustomZipFl(zipfile.ZipFile):

    def safe_extract(self, path=None, members=None):
        self.extractall(path, members)


class CustomTarFl(TarFile):

    def write(self, name, arcname=None):
        """Clone ``self.add`` method to the new name."""
        return self.add(name, arcname)

    def safe_extract(self, path: str | Path = ".", members=None):
        path: Path = path if isinstance(path, Path) else Path(path)
        # NOTE: For Python version >= 3.12
        if sys.version_info >= (3, 12):
            self.extractall(path, members, filter="data")
            return
        self.extractall(path, members)


class Dir:
    """Open File Object"""

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: str,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        _compress, sub = must_split(compress, ":", maxsplit=1)
        self.compress: DirCompressType = _compress
        self.sub_compress: str = sub or "_"

        # NOTE: Action anything after set up attributes.
        self.after_set_attrs()

    def after_set_attrs(self) -> None: ...

    def open(self, *, mode: str, **kwargs) -> OpenDirProtocol:
        """Open dir"""
        if self.compress in {"zip"}:
            ZIP_COMPRESS: dict[str, Any] = {
                "_": zipfile.ZIP_DEFLATED,
                "bz2": zipfile.ZIP_BZIP2,
            }

            return CustomZipFl(
                self.path,
                mode=mode,
                compression=ZIP_COMPRESS[self.sub_compress],
                **kwargs,
            )
        elif self.compress in {"tar"}:
            TAR_COMPRESS: dict[str, str] = {
                "_": "gz",
                "gz": "gz",
                "bz2": "bz2",
                "xz": "xz",
            }

            return CustomTarFl.open(
                self.path,
                mode=f"{mode}:{TAR_COMPRESS[self.sub_compress]}",
            )
        raise NotImplementedError


class EnvFl(Fl):
    """Env object which mapping search engine"""

    keep_newline: ClassVar[bool] = False
    default: ClassVar[str] = ""

    def read(self, *, update: bool = True) -> dict[str, str]:
        with self.open(mode="r") as _r:
            _r.seek(0)
            _result: dict = search_env(
                _r.read(),
                keep_newline=self.keep_newline,
                default=self.default,
            )
            if update:
                os.environ.update(**_result)
            return _result

    def write(self, data: dict[str, Any]) -> None:
        raise NotImplementedError


class YamlFl(Fl):
    """Yaml File Object

    .. noted::
        - The boolean value in the yaml file
            - true: Y, true, Yes, ON
            - false: n, false, No, off
    """

    def read(self, safe: bool = True) -> dict[str, Any]:
        with self.open(mode="r") as _r:
            return yaml.load(_r.read(), (SafeLoader if safe else UnsafeLoader))

    def write(self, data: dict[str, Any]) -> None:
        with self.open(mode="w") as _w:
            yaml.dump(data, _w, default_flow_style=False)


class YamlFlResolve(YamlFl):

    def read(self, safe: bool = True) -> dict[str, Any]:
        """Reading Yaml data with does not convert boolean value.
        Note:
            Handle top level yaml property ``on``
            docs: https://github.com/yaml/pyyaml/issues/696

            import re
            from yaml.resolver import Resolver

            # zap the Resolver class' internal dispatch table
            Resolver.yaml_implicit_resolvers = {}

            # NOTE: Current Resolver
            Resolver.add_implicit_resolver(
                    'tag:yaml.org,2002:bool',
                    re.compile(r'''^(?:yes|Yes|YES|no|No|NO
                                |true|True|TRUE|false|False|FALSE
                                |on|On|ON|off|Off|OFF)$''', re.X),
                    list('yYnNtTfFoO'))

            # NOTE: The 1.2 bool impl Resolver:
            Resolver.add_implicit_resolver(
                    'tag:yaml.org,2002:bool',
                    re.compile(r'^(?:true|false)$', re.X),
                    list('tf'))
        """
        from yaml.resolver import Resolver

        revert = Resolver.yaml_implicit_resolvers.copy()

        for ch in "OoYyNn":
            if len(Resolver.yaml_implicit_resolvers[ch]) == 1:
                del Resolver.yaml_implicit_resolvers[ch]
            else:
                Resolver.yaml_implicit_resolvers[ch] = [
                    x
                    for x in Resolver.yaml_implicit_resolvers[ch]
                    if x[0] != "tag:yaml.org,2002:bool"
                ]

        with self.open(mode="r") as _r:
            rs: dict[str, Any] = yaml.load(
                _r.read(), (SafeLoader if safe else UnsafeLoader)
            )
            # NOTE: revert resolver when want to use safe load.
            Resolver.yaml_implicit_resolvers = revert
            return rs


class YamlEnvFl(YamlFl):
    """Yaml object which mapping search environment variable."""

    raise_if_not_default: ClassVar[bool] = False
    default: ClassVar[str] = "null"
    escape: ClassVar[str] = "<ESCAPE>"

    @staticmethod
    def prepare(x: str) -> str:
        return x

    def read(self, safe: bool = True) -> dict[str, Any]:
        with self.open(mode="r") as _r:
            _env_replace: str = search_env_replace(
                yaml.dump(yaml.load(_r.read(), UnsafeLoader)),
                raise_if_default_not_exists=self.raise_if_not_default,
                default=self.default,
                escape=self.escape,
                caller=self.prepare,
            )
            if _result := yaml.load(
                _env_replace,
                (SafeLoader if safe else UnsafeLoader),
            ):
                return _result
            return {}

    def write(self, data: dict[str, Any]) -> None:
        raise NotImplementedError


class CsvFl(Fl):
    def read(self) -> list[str]:
        with self.open(mode="r") as _r:
            try:
                dialect = csv.Sniffer().sniff(_r.read(128))
                _r.seek(0)
                return list(csv.DictReader(_r, dialect=dialect))
            except csv.Error:
                return []

    def write(
        self,
        data: Union[list[Any], dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> None:
        mode = mode or "w"
        assert mode in (
            "a",
            "w",
        ), "save mode must contain only value `a` nor `w`."
        with self.open(mode=mode, newline="") as _w:
            _has_data: bool = True
            if isinstance(data, dict):
                data: list = [data]
            elif not data:
                data: list = [{}]
                _has_data: bool = False
            if _has_data:
                writer = csv.DictWriter(
                    _w,
                    fieldnames=list(data[0].keys()),
                    lineterminator="\n",
                    **kwargs,
                )
                if mode == "w" or not self.has_header:
                    writer.writeheader()
                writer.writerows(data)

    @property
    def has_header(self) -> bool:
        with self.open(mode="r") as _r:
            try:
                return csv.Sniffer().has_header(_r.read(128))
            except csv.Error:
                return False


class CsvPipeFl(CsvFl):
    def after_set_attrs(self) -> None:
        csv.register_dialect(
            "pipe_delimiter", delimiter="|", quoting=csv.QUOTE_ALL
        )

    def read(self) -> list:
        with self.open(mode="r") as _r:
            try:
                return list(
                    csv.DictReader(_r, delimiter="|", quoting=csv.QUOTE_ALL)
                )
            except csv.Error:
                return []

    def write(
        self,
        data: Union[list[Any], dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> None:
        mode = mode or "w"
        assert mode in {
            "a",
            "w",
        }, "save mode must contain only value `a` nor `w`."
        with self.open(mode=mode, newline="") as _w:
            _has_data: bool = True
            if isinstance(data, dict):
                data: list = [data]
            elif not data:
                data: list = [{}]
                _has_data: bool = False
            if _has_data:
                writer = csv.DictWriter(
                    _w,
                    fieldnames=list(data[0].keys()),
                    lineterminator="\n",
                    delimiter="|",
                    quoting=csv.QUOTE_ALL,
                    **kwargs,
                )
                if mode == "w" or not self.has_header:
                    writer.writeheader()
                writer.writerows(data)


class JsonFl(Fl):
    def read(self) -> Union[dict[Any, Any], list[Any]]:
        with self.open(mode="r") as _r:
            try:
                return json.loads(_r.read())
            except json.decoder.JSONDecodeError:
                return {}

    def write(
        self,
        data,
        *,
        indent: int = 4,
    ) -> None:
        _w: IO
        with self.open(mode="w") as _w:
            if self.compress:
                _w.write(json.dumps(data))
            else:
                json.dump(data, _w, indent=indent)


class JsonEnvFl(JsonFl):
    raise_if_not_default: bool = False
    default: str = "null"
    escape: str = "<ESCAPE>"

    @staticmethod
    def prepare(x: str) -> str:
        return x

    def read(self) -> Union[dict[Any, Any], list[Any]]:
        with self.open(mode="rt") as _r:
            return json.loads(
                search_env_replace(
                    _r.read(),
                    raise_if_default_not_exists=self.raise_if_not_default,
                    default=self.default,
                    escape=self.escape,
                    caller=self.prepare,
                )
            )

    def write(self, data, *, indent: int = 4) -> None:
        raise NotImplementedError


class TomlFl(Fl):
    def read(self):
        with self.open(mode="rt") as _r:
            return toml.loads(_r.read())

    def write(self, data: Any) -> None:
        with self.open(mode="wt") as _w:
            toml.dump(data, _w)


class TomlEnvFl(TomlFl):
    raise_if_not_default: bool = False
    default: str = "null"
    escape: str = "<ESCAPE>"

    @staticmethod
    def prepare(x: str) -> str:
        return x

    def read(self):
        with self.open(mode="rt") as _r:
            return toml.loads(
                search_env_replace(
                    _r.read(),
                    raise_if_default_not_exists=self.raise_if_not_default,
                    default=self.default,
                    escape=self.escape,
                    caller=self.prepare,
                )
            )


class PickleFl(Fl):
    def read(self):
        with self.open(mode="rb") as _r:
            return pickle.loads(_r.read())

    def write(self, data):
        with self.open(mode="wb") as _w:
            pickle.dump(data, _w)


class MarshalFl(Fl):
    def read(self):
        with self.open(mode="rb") as _r:
            return marshal.loads(_r.read())

    def write(self, data):
        with self.open(mode="wb") as _w:
            marshal.dump(data, _w)


class MsgpackFl(Fl):
    def read(self):
        with self.open(mode="rb") as _r:
            return msgpack.loads(_r.read())

    def write(self, data):
        with self.open(mode="wb") as _w:
            msgpack.dump(data, _w)


__all__ = (
    "Fl",
    "EnvFl",
    "JsonFl",
    "JsonEnvFl",
    "YamlFl",
    "YamlFlResolve",
    "YamlEnvFl",
    "CsvFl",
    "CsvPipeFl",
    "TomlFl",
    "TomlEnvFl",
    "MarshalFl",
    "MsgpackFl",
    "PickleFl",
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Callable, TypeVar

from ddeutil.core import import_string, str2args

try:
    from .__base import RegexConf
except ImportError:  # no cove
    from __base import RegexConf


T = TypeVar("T")


def map_secret(value: T, secrets: dict[str, str]) -> T:
    """Map the secret value to a any input data.

    :param value: A data that want to map secrets
    :param secrets: A mapping of secrets
    :type secrets: dict[str, str]

    Examples:
        >>> map_secret(
        ...     "Value include secrets: s3://@secrets{foo}",
        ...     secrets={"foo": "bar"},
        ... )
        'Value include secrets: s3://bar'
    """
    if isinstance(value, dict):
        return {k: map_secret(value[k], secrets) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([map_secret(i, secrets) for i in value])
    elif not isinstance(value, str):
        return value
    for search in RegexConf.RE_SECRETS.finditer(value):
        searches: dict = search.groupdict()
        if "." in (br := searches["braced"]):
            raise ValueError(
                f"The @secrets: {br!r}, should not contain dot ('.') char"
            )
        value: str = value.replace(
            searches["search"],
            secrets.get(br.strip(), searches["braced_default"]),
        )
    return value


def map_importer(value: T) -> T:
    """Map the function result to configuration data.

    :param value: A data that want to map imported function with arguments.

    Examples:
        >>> map_importer(
        ...     "Test @function{ddeutil.io.__base.add_newline:'a',newline='|'}"
        ... )
        'Test a|'
    """
    if isinstance(value, dict):
        return {k: map_importer(value[k]) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([map_importer(i) for i in value])
    elif not isinstance(value, str):
        return value
    for search in RegexConf.RE_FUNCTION.finditer(value):
        searches: dict = search.groupdict()
        if not callable(_fn := import_string(searches["function"])):
            raise ValueError(
                f'The @function: {searches["function"]!r} is not callable.',
            )
        args, kwargs = str2args(searches["arguments"])
        value: str = value.replace(searches["search"], _fn(*args, **kwargs))
    return value


def map_func(value: T, fn: Callable[[str], str]) -> T:
    """Map any function from input argument to configuration data.

    Examples:
        >>> map_func({"foo": "bar"}, lambda x: x + "!")
        {'foo': 'bar!'}
        >>> map_func(("foo", "bar"), lambda x: x + "!")
        ('foo!', 'bar!')
    """
    if isinstance(value, dict):
        return {k: map_func(value[k], fn) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([map_func(i, fn) for i in value])
    elif not isinstance(value, str):
        return value
    return fn(value)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Annotated
from urllib.parse import unquote_plus

from pydantic import EncodedStr, EncoderProtocol
from pydantic.networks import UrlConstraints
from pydantic_core import Url

CustomUrl = Annotated[
    Url,
    UrlConstraints(
        host_required=True,
        default_port=1234,
    ),
]

LocalUrl = Annotated[
    Url,
    UrlConstraints(
        default_host="127.0.0.1",
        default_port=None,
    ),
]


class UnquoteEncoder(EncoderProtocol):
    @classmethod
    def decode(cls, data: bytes) -> bytes:
        # NOTE:
        #   We have to use unquote rather than unquote_plus because only unquote
        #   can work with bytes objects. This may be a limitation if your URL
        #   string contains encoded spaces.
        return str.encode(unquote_plus(data))


EncodedStr = Annotated[str, EncodedStr(encoder=UnquoteEncoder)]



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import (
    Annotated,
    Optional,
)

from pydantic import BaseModel, Field


class Const(BaseModel):
    """Constraint Model

    Examples:
        >>> Const(of='foo').name
        'foo_const'
    """

    of: Annotated[
        Optional[str],
        Field(description="Owner of Constraint"),
    ] = None

    @property
    def name(self) -> str:
        if not self.of:
            raise ValueError(
                "This constraint does not pass `of` value for take ownership."
            )
        return f"{self.of}_const"


class Pk(Const):
    """Primary Key Model.

    Examples:
        >>> Pk(of="foo", cols=["bar", "baz"]).name
        'foo_bar_baz_pk'
    """

    cols: Annotated[
        list[str],
        Field(default_factory=list, description="List of primary key columns"),
    ]

    @property
    def name(self) -> str:
        if not self.of:
            raise ValueError(
                "This constraint does not pass `of` value for take ownership."
            )
        if self.cols:
            return f'{self.of}_{"_".join(self.cols)}_pk'
        raise ValueError("This primary key does not have any columns.")


class Ref(BaseModel):
    """Reference Model

    Examples:
        >>> data = {
        ...     "tbl": "foo",
        ...     "col": "bar",
        ... }
        >>> Ref.model_validate(data).tbl
        'foo'
    """

    tbl: str
    col: str


class Fk(Const):
    """Foreign Key Model.

    Examples:
        >>> data = {
        ...     "of": "foo",
        ...     "to": "bar",
        ...     "ref": {
        ...         "tbl": "ref_table",
        ...         "col": "ref_column"
        ...     }
        ... }
        >>> Fk.model_validate(data).name
        'foo_bar_ref_table_ref_column_fk'
    """

    to: str
    ref: Ref

    @property
    def name(self) -> str:
        if not self.of:
            raise ValueError(
                "This constraint does not pass `of` value for take ownership."
            )
        return f"{self.of}_{self.to}_{self.ref.tbl}_{self.ref.col}_fk"



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import (
    Annotated,
    Any,
    Literal,
    Optional,
    Union,
)

from pydantic import Field, ValidationInfo
from pydantic.functional_validators import field_validator, model_validator
from pydantic.types import SecretStr
from pydantic_core import Url

from .__base import BaseUpdatableModel
from .__types import CustomUrl
from .utils import unquote_str


class BaseConn(BaseUpdatableModel):
    type: str = "base"


class Conn(BaseConn):
    type: Literal["conn"] = "conn"
    dialect: str
    host: Optional[str] = None
    port: Optional[int] = None
    user: Optional[str] = None
    pwd: Optional[SecretStr] = None
    endpoint: Optional[str] = None
    options: Annotated[dict[str, Any], Field(default_factory=dict)]

    @classmethod
    def from_url(cls, url: Union[Url, str]) -> Conn:
        if isinstance(url, str):
            url = Url(url=url)
        elif not isinstance(url, Url):
            raise ValueError("A url value must be string or `Url` object")
        return cls(
            dialect=url.scheme,
            host=unquote_str(url.host),
            port=url.port,
            user=unquote_str(url.username),
            pwd=unquote_str(url.password),
            endpoint=url.path,
            options=dict(url.query_params()),
        )

    @field_validator("endpoint")
    def _remove_slash_on_endpoint(cls, v: str) -> str:
        return v[1:] if v.startswith("/") else v


class FlConn(BaseConn):
    type: Literal["file"] = "file"
    sys: str
    pointer: Optional[str] = None
    port: Optional[int] = None
    user: Optional[str] = None
    pwd: Optional[SecretStr] = None
    path: str
    options: Annotated[dict[str, Any], Field(default_factory=dict)]

    @classmethod
    def from_url(cls, url: Union[Url, str]) -> FlConn:
        if isinstance(url, str):
            url = Url(url=url)
        elif not isinstance(url, Url):
            raise ValueError("A url value must be string or `FileUrl` object")
        return cls(
            sys=url.scheme,
            pointer=unquote_str(url.password),
            port=url.port,
            user=unquote_str(url.username),
            pwd=unquote_str(url.password),
            path=url.path,
            options=dict(url.query_params()),
        )

    @field_validator("path")
    def _remove_slash_on_endpoint(cls, v: str) -> str:
        return v[1:] if v.startswith("/") else v


class DbConn(BaseConn):
    """Database Connection Model

    Example:
        *   {
                "type": "db",
                "driver": "sqlite",
            }
    """

    type: Literal["db"] = "db"
    driver: str
    host: str
    port: int
    user: str
    pwd: SecretStr
    db: str
    options: Annotated[dict[str, Any], Field(default_factory=dict)]

    @classmethod
    def from_url(cls, url: Union[CustomUrl, str]) -> DbConn:
        if isinstance(url, str):
            url = CustomUrl(url=url)
        elif not isinstance(url, CustomUrl):
            raise ValueError("A url value must be string or `CustomUrl` object")
        return cls(
            driver=url.scheme,
            host=unquote_str(url.host),
            port=url.port,
            user=unquote_str(url.username),
            pwd=unquote_str(url.password),
            db=url.path,
            options=dict(url.query_params()),
        )

    @field_validator("db")
    def _remove_slash_on_db(cls, v: str, info: ValidationInfo, **kwargs) -> str:
        return v[1:] if v.startswith("/") else v

    @model_validator(mode="after")
    def _add_path_to_options(self):
        if len(dbs := self.db.replace("\\", "/").split("/")) > 1:
            self.db: str = dbs[0]
            self.options["_path"] = "/".join(dbs[1:])


# @remove_validators('_check_db_name')
class DbWithSlashConn(DbConn):
    """"""

    @field_validator("db")
    def _remove_slash_on_db(cls, v: str) -> str:
        return v

    @model_validator(mode="after")
    def _add_path_to_options(self): ...



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import abc
import contextlib
import inspect
import json
import logging
import shutil
import sqlite3
from collections.abc import Generator, Iterator
from datetime import datetime
from pathlib import Path
from sqlite3 import Connection
from typing import Any, Union

from .__base import (
    Fl,
    JsonFl,
    PathSearch,
    YamlEnvFl,
    rm,
)
from .exceptions import ConfigArgumentError

DEFAULT_OPEN_FILE: type[Fl] = YamlEnvFl
DEFAULT_OPEN_FILE_STG: type[Fl] = JsonFl
DEFAULT_EXCLUDED_FMT: tuple[str, ...] = (".json", ".toml")


class ConfABC(abc.ABC):
    """Config Adapter abstract object."""

    @abc.abstractmethod
    def load_stage(self, name: str) -> dict:
        raise NotImplementedError

    @abc.abstractmethod
    def save_stage(self, name: str, data: dict, merge: bool = False) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def remove_stage(self, name: str, data_name: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def create(self, name: str, **kwargs) -> None:
        raise NotImplementedError


class BaseConfFl:
    """Base Config File object for getting data with `.yaml` format and mapping
    environment variables to the content data.
    """

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: str | None = None,
        open_file: type[Fl] | None = None,
        excluded_fmt: tuple[str, ...] | None = None,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        self.compress: str | None = compress
        self.open_file: type[Fl] = open_file or DEFAULT_OPEN_FILE
        self.excluded_fmt: tuple[str] = excluded_fmt or DEFAULT_EXCLUDED_FMT
        if not self.path.exists():
            self.path.mkdir(parents=True)

    def load(self, name: str, *, order: int = 1) -> dict[str, Any]:
        """Return configuration data from name of the config.

        :param name: A name of config key that want to search in the path.
        :type name: str
        :param order: An order number that want to get from ordered list
            of duplicate data.
        :type order: int(=1)

        :rtype: dict[str, Any]
        :returns: A loaded data from open file object that already adding
            `alias` key with a config name.
        """
        rs: list[dict[Any, Any]]
        if rs := [
            {"alias": name} | data
            for file in self.files(excluded=self.excluded_fmt)
            if (
                data := self.open_file(path=file, compress=self.compress)
                .read()
                .get(name)
            )
        ]:
            try:
                if order > len(rs):
                    raise IndexError
                return sorted(
                    rs,
                    key=lambda x: (
                        datetime.fromisoformat(x.get("version", "1990-01-01")),
                        len(x),
                    ),
                    reverse=False,
                )[-order]
            except IndexError:
                logging.warning(
                    f"Does not load config {name!r} with order: -{order}"
                )
        return {}

    def files(
        self,
        path: str | None = None,
        name: str | None = None,
        *,
        excluded: list[str] | None = None,
    ) -> Iterator[Path]:
        """Return all files that exists in the loading path."""
        yield from filter(
            lambda x: x.is_file(),
            (
                PathSearch(root=(path or self.path), exclude=excluded).pick(
                    filename=(name or "*")
                )
            ),
        )

    def move(self, path: Path, dest: Path) -> None:
        """Copy filename to destination path."""
        if not dest.parent.exists():
            dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(self.path / path, dest)


class ConfFl(BaseConfFl, ConfABC):
    """Config File Loading Object for get data from configuration and stage.

    :param path: A path of files to action.
    :type path: str | Path
    :param compress: str | None : A compress type of action file.
    """

    def __init__(
        self,
        path: str | Path,
        *,
        compress: str | None = None,
        open_file: type[Fl] | None = None,
        excluded_fmt: list[str] | None = None,
        open_file_stg: type[Fl] | None = None,
    ):
        """Main initialize of config file loading object."""
        super().__init__(
            path,
            compress=compress,
            open_file=open_file,
            excluded_fmt=excluded_fmt,
        )
        self.open_file_stg: type[Fl] = open_file_stg or DEFAULT_OPEN_FILE_STG

    def load_stage(
        self,
        path: str | Path,
        *,
        default: Any = None,
    ) -> Union[dict[Any, Any], list[Any]]:
        """Return content data from file with filename, default empty dict."""
        try:
            return self.open_file_stg(
                path=path,
                compress=self.compress,
            ).read()
        except FileNotFoundError:
            return default if (default is not None) else {}

    def save_stage(
        self,
        path: str | Path,
        data: Union[dict[Any, Any], list[Any]],
        *,
        merge: bool = False,
    ) -> None:
        """Write content data to file with filename. If merge is true, it will
        load current data from file and merge the data content together
        before write.
        """
        if not merge:
            self.open_file_stg(path, compress=self.compress).write(data)
            logging.debug(f"Start writing data to {path}")
            return
        elif merge and (
            "mode"
            in inspect.getfullargspec(self.open_file_stg.write).annotations
        ):
            self.open_file_stg(path, compress=self.compress).write(
                **{
                    "data": data,
                    "mode": "a",
                }
            )
            return

        all_data: Union[dict, list] = self.load_stage(path=path)
        try:
            if isinstance(all_data, list):
                _merge_data: Union[dict, list] = all_data
                if isinstance(data, dict):
                    _merge_data.append(data)
                else:
                    _merge_data.extend(data)
            else:
                _merge_data: dict = all_data | data
            self.open_file_stg(path, compress=self.compress).write(_merge_data)
        except TypeError as err:
            rm(path=path)
            if all_data:
                self.open_file_stg(path, compress=self.compress).write(
                    all_data,
                )
            raise err

    def remove_stage(self, path: str, name: str) -> None:
        """Remove data by name insided the staging file with filename."""
        if all_data := self.load_stage(path=path):
            all_data.pop(name, None)
            self.open_file_stg(path, compress=self.compress).write(
                all_data,
            )

    def create(
        self,
        path: Path,
        *,
        initial_data: Any = None,
    ) -> None:
        """Create filename in path."""
        if not path.exists():
            self.save_stage(
                path=path,
                data=({} if initial_data is None else initial_data),
                merge=False,
            )


class BaseConfSQLite:
    """Base Config SQLite object for getting data with SQLite database from
    file storage."""

    def __init__(
        self,
        path: Union[str, Path],
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        if not self.path.exists():
            self.path.mkdir(parents=True)

    @contextlib.contextmanager
    def connect(self, database: str) -> Generator[Connection, None, None]:
        """Return SQLite Connection context."""
        _conn: Connection = sqlite3.connect(self.path / database, timeout=3)
        _conn.row_factory = self.dict_factory
        try:
            yield _conn
        except sqlite3.Error as err:
            logging.error(err)
            raise ConfigArgumentError(
                "syntax", f"SQLite syntax error {err}"
            ) from err
        _conn.commit()
        _conn.close()

    @staticmethod
    def dict_factory(cursor, row) -> dict[str, Any]:
        """Result of dictionary factory.

        :note:
            Another logic of dict factory.

            - dict(
                [
                    (col[0], row[idx])
                    for idx, col in enumerate(cursor.description)
                ]
            )

            - dict(zip([col[0] for col in cursor.description], row))
        """
        return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


class ConfSQLite(BaseConfSQLite, ConfABC):
    """Config SQLite Loading Object for get data from configuration and save
    stage data to the one table.
    """

    def load_stage(
        self,
        table: str,
        default: dict[Any, Any] | None = None,
    ) -> dict[Any, Any]:
        """Return content data from database with table name, default empty
        dict."""
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            cur.execute(f"select * from {_table};")
            if result := cur.fetchall():
                return {_["name"]: self.convert_type(_) for _ in result}
            return default if (default is not None) else {}

    def save_stage(
        self,
        table: str,
        data: dict[str, Any],
        merge: bool = False,
    ) -> None:
        """Write content data to database with table name. If merge is true, it
        will update or insert the data content.
        """
        _db, _table = table.rsplit("/", maxsplit=1)
        _data: dict = self.prepare_values(data.get(list(data.keys())[0]))
        with self.connect(_db) as conn:
            cur = conn.cursor()
            doing: str = "nothing"
            if merge:
                _doing_list = [
                    f"{_} = excluded.{_}" for _ in _data if _ != "name"
                ]
                doing: str = f'update set {", ".join(_doing_list)}'
            query: str = (
                f'insert into {_table} ({", ".join(_data.keys())}) values '
                f'({":" + ", :".join(_data.keys())}) '
                f"on conflict ( name ) do {doing};"
            )
            cur.execute(query, _data)

    def remove_stage(
        self,
        table: str,
        data_name: str,
    ) -> None:
        """Remove data by name from table in database with table name."""
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            cur.execute(f"delete from {_table} where name = '{data_name}';")

    def create(
        self,
        table: str,
        schemas: dict[str, str] | None = None,
    ) -> None:
        """Create table in database."""
        if not schemas:
            raise ConfigArgumentError(
                "schemas",
                (
                    f"in `create` method of {self.__class__.__name__} "
                    f"was required"
                ),
            )
        _schemas: str = ", ".join([f"{k} {v}" for k, v in schemas.items()])
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            cur.execute(f"create table if not exists {_table} ({_schemas})")

    @staticmethod
    def prepare_values(
        values: dict[str, Union[str, int, float]],
    ) -> dict[str, Union[str, int, float]]:
        """Return prepare value with dictionary type to string
        to source system.
        """
        rs: dict[str, Union[str, int, float]] = values.copy()
        for _ in values:
            if isinstance(values[_], dict):
                rs[_] = json.dumps(values[_])
        return rs

    @staticmethod
    def convert_type(
        data: dict[str, Union[str, int, float]],
        key: str | None = None,
    ) -> dict[str, Any]:
        """Return converted value from string to dictionary
        from source system.
        """
        _key: str = key or "data"
        rs: dict[str, Any] = data.copy()
        rs[_key] = json.loads(data[_key])
        return rs


__all__ = (
    "ConfABC",
    "ConfFl",
    "ConfSQLite",
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import re
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Optional,
    Union,
)

from pydantic import BaseModel, Field, ValidationInfo
from pydantic.functional_validators import field_validator

from .__base import YamlEnvFl
from .__conf import UPDATE_KEY, VERSION_KEY
from .exceptions import ConfigArgumentError

TupleStr = tuple[str, ...]


FMT_NAMES: TupleStr = (
    "naming",
    "domain",
    "environ",
    "timestamp",
    "version",
    "compress",
    "extension",
)

RULE_FIX: TupleStr = (
    "timestamp",
    "version",
    "compress",
)


class Rule(BaseModel):
    """Rule Model that keep rule setting data for Register object.

    Examples:
        >>> rule = {
        ...     "timestamp": {"minutes": 15},
        ...     "excluded": [],
        ...     "compress": None,
        ... }
    """

    timestamp: Optional[dict[str, int]] = Field(default_factory=dict)
    version: Annotated[Optional[str], Field()] = None
    excluded: Optional[list[str]] = Field(default_factory=list)
    compress: Annotated[
        Optional[str],
        Field(description="Compress type"),
    ] = None


class Stage(BaseModel):
    """Stage Model that keep stage data for transition data.

    Examples:
        >>> stage = {
        ...     "raw": {
        ...         "format": "",
        ...         "rules": {},
        ...     },
        ... }
    """

    alias: str
    format: str
    rules: Rule = Field(default_factory=Rule, description="Rule of stage")
    layer: int = Field(default=0)

    @field_validator("format", mode="before")
    def validate_format(cls, value: str, info: ValidationInfo):
        # NOTE:
        #   Validate the name in format string should contain any format name.
        if not (
            _searches := re.findall(
                r"{(?P<name>\w+):?(?P<format>[^{}]+)?}",
                value,
            )
        ):
            raise ConfigArgumentError(
                "format",
                (
                    f'This `{info.data["alias"]}` stage format dose not '
                    f"include any format name, the stage file was duplicated."
                ),
            )

        # NOTE: Validate the name in format string should exist in `FMT_NAMES`.
        if any((_search[0] not in FMT_NAMES) for _search in _searches):
            raise ConfigArgumentError(
                "format",
                "This stage have an unsupported format name.",
            )
        return value

    @field_validator("format", mode="after")
    def validate_rule_relate_with_format(cls, value, info: ValidationInfo):
        # NOTE: Validate a format of stage that relate with rules.
        for validator in RULE_FIX:
            if getattr(info.data.get("rules", {}), validator, None) and (
                validator not in value
            ):
                raise ConfigArgumentError(
                    (
                        "format",
                        validator,
                    ),
                    (
                        f"This stage set `{validator}` rule but does not have "
                        f"a `{validator}` format name in the format."
                    ),
                )
        return value


class PathData(BaseModel):
    """Path Data Model that keep necessary paths for register or loading object.

    Examples:
        >>> path_data = {
        ...     "root": "./",
        ...     "data": "./data",
        ...     "conf": "./config",
        ... }
    """

    root: Path = Field(default_factory=Path)
    data: Path = Field(default=None, validate_default=True)
    conf: Path = Field(default=None, validate_default=True)

    @field_validator("root", mode="before")
    def prepare_root(cls, v: Union[str, Path]) -> Path:
        return Path(v) if isinstance(v, str) else v

    @field_validator("data", "conf", mode="before")
    def prepare_path_from_path_str(cls, v, info: ValidationInfo) -> Path:
        if v is not None:
            return Path(v) if isinstance(v, str) else v
        return info.data["root"] / info.field_name


class Value(BaseModel):
    dt_fmt: str = Field(default="%Y-%m-%d %H:%M:%S")
    excluded: TupleStr = Field(default=(VERSION_KEY, UPDATE_KEY))


class Engine(BaseModel):
    paths: PathData = Field(default_factory=PathData)
    values: Value = Field(default_factory=Value)


class Params(BaseModel, validate_assignment=True):
    stages: dict[str, Stage] = Field(default_factory=dict)
    engine: Engine = Field(default_factory=Engine)

    @classmethod
    def from_yaml(cls, path: Union[str, Path]):
        """Load params from .yaml file"""
        return cls.model_validate(YamlEnvFl(path).read())

    @field_validator("stages", mode="before")
    def prepare_order_layer(cls, value: dict[str, dict[Any, Any]]):
        for i, k in enumerate(value, start=1):
            value[k] = value[k].copy() | {"layer": i, "alias": k}
        return value

    @property
    def stage_final(self) -> str:
        """Return the final stage name that ordered from layer value."""
        return max(self.stages.items(), key=lambda i: i[1].layer)[0]

    @property
    def stage_first(self) -> str:
        """Return the first stage name that ordered from layer value which
        does not be the base stage.
        """
        return min(self.stages.items(), key=lambda i: i[1].layer)[0]

    def get_stage(self, name: str) -> Stage:
        """Return Stage model that match with stage name. If an input stage
        value equal 'base', it will return the default stage model.

        :param name: A stage name that want to getting from this params.
        :type name: str
        """
        if name == "base":
            return Stage.model_validate(
                {
                    "format": "{naming}_{timestamp}",
                    "alias": "base",
                }
            )
        elif name not in self.stages:
            raise ConfigArgumentError(
                "stage",
                f"Cannot get stage: {name!r} cause it does not exists",
            )
        return self.stages[name].model_copy()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import logging
from collections.abc import Container
from typing import (
    AbstractSet,
    Any,
    Callable,
    TypeVar,
    Union,
)

from pydantic import BaseModel, ConfigDict, ValidationError
from typing_extensions import Self

T = TypeVar("T")
TModel = TypeVar("TModel", bound=BaseModel)

AbstractSetOrDict = Union[
    AbstractSet[Union[int, str]],
    dict[Union[int, str], Any],
]


class __BaseModel(BaseModel):
    # NOTE:
    #   This config allow to validate before assign new data to any field
    model_config = ConfigDict(
        validate_assignment=True,
        use_enum_values=True,
        populate_by_name=True,
    )


class BaseUpdatableModel(__BaseModel):
    """Base Model that was implemented updatable method and properties."""

    @classmethod
    def get_field_names(cls, alias=False):
        return list(cls.model_json_schema(alias).get("properties").keys())

    @classmethod
    def get_properties(cls) -> list:
        """Return list of properties of this model"""
        return [
            prop
            for prop in cls.__dict__
            if isinstance(cls.__dict__[prop], property)
        ]

    def dict(
        self,
        *,
        include: AbstractSetOrDict = None,
        exclude: AbstractSetOrDict = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        **kwargs,
    ) -> dict[str, Any]:
        """Override the dict function to include our properties
        docs: https://github.com/pydantic/pydantic/issues/935
        """
        attribs = super().model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            **kwargs,
        )
        props: list = self.get_properties()

        # Include and exclude properties
        if include:
            props: list = [prop for prop in props if prop in include]
        if exclude:
            props: list = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})
        return attribs

    def update(self, data: dict) -> Self:
        """Updatable method for update data to existing model data.
        docs: https://github.com/pydantic/pydantic/discussions/3139
        """
        update = self.dict()
        update.update(data)
        for k, v in (
            self.model_validate(update).dict(exclude_defaults=True).items()
        ):
            logging.debug(
                f"Updating value '{k}' from '{getattr(self, k, None)}' to '{v}'"
            )
            setattr(self, k, v)
        return self


def invalid_to_none(v: Any, handler: Callable[[Any], Any]) -> Any:
    """Invalid to None value function.

    Examples:
        >>> from typing import Annotated, Optional
        >>> from pydantic import WrapValidator
        >>> class Foo(BaseModel):
        ...     age: Annotated[Optional[int], WrapValidator(invalid_to_none)]
        ...     name: Optional[str]
        >>> Foo(age="invalid", name="Jim").model_dump()
        {'age': None, 'name': 'Jim'}

    Examples:
        We able to implement this to all fields on any sub-classes.
        >>> from typing import get_args, get_origin
        >>> class CustomBaseModel(BaseModel):
        ...     def __init_subclass__(cls, **kwargs: Any) -> None:
        ...         for name, annotation in cls.__annotations__.items():
        ...             # exclude protected/private attributes
        ...             if name.startswith("_"):
        ...                 continue
        ...             validator = WrapValidator(invalid_to_none)
        ...             if get_origin(annotation) is Annotated:
        ...                 cls.__annotations__[name] = Annotated[
        ...                     (*get_args(annotation), ),
        ...                     validator,
        ...                 ]
        ...             else:
        ...                 cls.__annotations__[name] = Annotated[
        ...                     annotation, validator
        ...                 ]
        >>> class Foo(CustomBaseModel):
        ...     age: Optional[int]
        ...     name: Optional[str]
        >>> Foo(age="invalid", name="Jim").model_dump()
        {'age': None, 'name': 'Jim'}

    Reference:
        *   https://stackoverflow.com/questions/76669927/ -
            pydantic-how-to-ignore-invalid-values-when-creating-model-instance
    """
    try:
        return handler(v)
    except ValidationError:
        return None


def _remove_validators(validators: T, names: Container[str]) -> T:
    return [v for v in validators if v.__name__ not in names]


def remove_validators(*names: str) -> Callable[[type[TModel]], type[TModel]]:
    def decorator(model: type[TModel]) -> type[TModel]:
        for name in names:
            if not getattr(model, name):
                raise ValueError(
                    f"{model.__name__} does not implement {name} validator"
                )
            model.__pydantic_decorators__.field_validators.pop(name)
        model.model_rebuild()
        return model

    return decorator
